"""Commands to finish up after a demultiplex run."""
import logging

import click

from cg.constants.constants import DRY_RUN
from cg.constants.demultiplexing import OPTION_BCL_CONVERTER
from cg.meta.demultiplex.demux_post_processing import DemuxPostProcessingAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


@click.group(name="finish")
def finish_group():
    """Finish up after demultiplexing."""


@finish_group.command(name="flow-cell")
@click.argument("flow-cell-directory-name")
@OPTION_BCL_CONVERTER
@click.option("--force", is_flag=True)
@DRY_RUN
@click.pass_obj
def finish_flow_cell(
    context: CGConfig, flow_cell_directory_name: str, bcl_converter: str, force: bool, dry_run: bool
):
    """Command to finish up a flow cell after demultiplexing.

    flow-cell-name is full flow cell name, e.g. '201203_D00483_0200_AHVKJCDRXX'.
    """
    demux_post_processing_api = DemuxPostProcessingAPI(config=context)
    demux_post_processing_api.set_dry_run(dry_run=dry_run)
    demux_post_processing_api.finish_flow_cell(
        flow_cell_directory_name=flow_cell_directory_name,
        bcl_converter=bcl_converter,
        force=force,
    )


@finish_group.command(name="all")
@click.pass_obj
@DRY_RUN
def finish_all_cmd(context: CGConfig, dry_run: bool):
    """Command to finish up all demultiplexed flow cells."""

    demux_post_processing_api = DemuxPostProcessingAPI(config=context)
    demux_post_processing_api.set_dry_run(dry_run=dry_run)
    is_error_raised: bool = demux_post_processing_api.finish_all_flow_cells()
    if is_error_raised:
        raise click.Abort



"""Test add customer CLI."""

from click.testing import CliRunner

from cg.cli.add import add
from cg.constants import EXIT_SUCCESS
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import Customer

NEW_TEST_CUST_NAME: str = "new_test_cust"
NEW_TEST_CUST_INTERNAL_ID: str = "new_test_cust_internal_id"


def test_add_customer(
    base_context: CGConfig, cli_runner: CliRunner, invoice_address: str, invoice_reference: str
):
    """Test add customer command."""
    # GIVEN database with some customers
    status_db: Store = base_context.status_db
    nr_customers: int = status_db._get_query(table=Customer).count()

    # WHEN adding a customer
    result = cli_runner.invoke(
        add,
        [
            "customer",
            "--invoice-address",
            invoice_address,
            "--invoice-reference",
            invoice_reference,
            NEW_TEST_CUST_INTERNAL_ID,
            NEW_TEST_CUST_NAME,
        ],
        obj=base_context,
    )

    # THEN exit successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN it should be stored in the database
    assert status_db._get_query(table=Customer).count() == nr_customers + 1


def test_add_customer_with_collaboration(
    base_context: CGConfig,
    cli_runner: CliRunner,
    collaboration_id: str,
    invoice_address: str,
    invoice_reference: str,
):
    """Test add customer with collaborator."""
    # GIVEN database with customers
    status_db: Store = base_context.status_db
    nr_customers: int = status_db._get_query(table=Customer).count()

    # WHEN adding a customer
    result = cli_runner.invoke(
        add,
        [
            "customer",
            "--invoice-address",
            invoice_address,
            "--invoice-reference",
            invoice_reference,
            "--collaboration-internal-ids",
            collaboration_id,
            NEW_TEST_CUST_INTERNAL_ID,
            NEW_TEST_CUST_NAME,
        ],
        obj=base_context,
    )

    new_customer: Customer = status_db.get_customer_by_internal_id(
        customer_internal_id=NEW_TEST_CUST_INTERNAL_ID
    )

    # THEN exit successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN it should be stored in the database
    assert status_db._get_query(table=Customer).count() == nr_customers + 1

    # THEN a collaborator should have been added
    assert new_customer.collaborations[0].internal_id == collaboration_id



import pytest

from cg.exc import OrderError
from cg.meta.orders.rnafusion_submitter import RnafusionSubmitter
from cg.models.orders.constants import OrderType
from cg.models.orders.order import OrderIn
from cg.store import Store


def test__validate_one_sample_per_case_multiple_samples(
    base_store: Store,
    rnafusion_order_to_submit: dict,
):
    """Tests the validation of an RNAFUSION order where two samples have the same family_name."""
    ### GIVEN an RNAFUSION order where the first and last sample share the same case
    order_data = OrderIn.parse_obj(obj=rnafusion_order_to_submit, project=OrderType.RNAFUSION)
    order_data.samples[-1].family_name = order_data.samples[0].family_name
    rnafusion_submitter = RnafusionSubmitter(status=base_store, lims=None)

    ### WHEN validating that each case has only one sample
    ### THEN an OrderError should be raised

    with pytest.raises(OrderError):
        rnafusion_submitter._validate_only_one_sample_per_case(order_data.samples)


def test__validate_one_sample_per_case_unique_samples(
    base_store: Store,
    rnafusion_order_to_submit: dict,
):
    """Tests the validation of an RNAFUSION order where all samples have unique family_name."""
    ### GIVEN an RNAFUSION order with unique family names
    order_data: OrderIn = OrderIn.parse_obj(
        obj=rnafusion_order_to_submit, project=OrderType.RNAFUSION
    )
    rnafusion_submitter: RnafusionSubmitter = RnafusionSubmitter(status=base_store, lims=None)

    ### WHEN validating that each case has only one sample
    rnafusion_submitter._validate_only_one_sample_per_case(order_data.samples)

    ### THEN no errors should be raised



import datetime as dt
import logging
import os
import shutil
from pathlib import Path
from subprocess import CalledProcessError

import click
from housekeeper.store.models import Bundle, Version

from cg.apps.environ import environ_email
from cg.constants import CASE_ACTIONS, EXIT_FAIL, EXIT_SUCCESS, Pipeline, Priority
from cg.constants.constants import AnalysisType, CaseActions, WorkflowManager
from cg.constants.priority import PRIORITY_TO_SLURM_QOS
from cg.exc import AnalysisNotReadyError, BundleAlreadyAddedError, CgDataError, CgError
from cg.meta.meta import MetaAPI
from cg.meta.workflow.fastq import FastqHandler
from cg.models.analysis import AnalysisModel
from cg.models.cg_config import CGConfig
from cg.store.models import Analysis, BedVersion, Case, CaseSample, Sample

LOG = logging.getLogger(__name__)


class AnalysisAPI(MetaAPI):
    """
    Parent class containing all methods that are either shared or overridden by other workflow APIs
    """

    def __init__(self, pipeline: Pipeline, config: CGConfig):
        super().__init__(config=config)
        self.pipeline = pipeline
        self._process = None

    @property
    def root(self):
        raise NotImplementedError

    @property
    def use_read_count_threshold(self) -> bool:
        """Defines whether the threshold for adequate read count should be passed for all samples
        when determining if the analysis for a case should be automatically started"""
        return False

    @property
    def process(self):
        raise NotImplementedError

    @property
    def fastq_handler(self):
        return FastqHandler

    @staticmethod
    def get_help(context):
        """
        If no argument is passed, print help text
        """
        if context.invoked_subcommand is None:
            click.echo(context.get_help())

    def verify_deliverables_file_exists(self, case_id: str) -> None:
        if not Path(self.get_deliverables_file_path(case_id=case_id)).exists():
            raise CgError(f"No deliverables file found for case {case_id}")

    def verify_case_config_file_exists(self, case_id: str) -> None:
        if not Path(self.get_case_config_path(case_id=case_id)).exists():
            raise CgError(f"No config file found for case {case_id}")

    def check_analysis_ongoing(self, case_id: str) -> None:
        if self.trailblazer_api.is_latest_analysis_ongoing(case_id=case_id):
            LOG.warning(f"{case_id} : analysis is still ongoing - skipping")
            raise CgError(f"Analysis still ongoing in Trailblazer for case {case_id}")

    def verify_case_path_exists(self, case_id: str) -> None:
        """Check if case path exists."""
        if not self.get_case_path(case_id=case_id).exists():
            LOG.info(f"No working directory for {case_id} exists")
            raise FileNotFoundError(f"No working directory for {case_id} exists")

    def get_priority_for_case(self, case_id: str) -> int:
        """Get priority from the status db case priority"""
        case_obj: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        return case_obj.priority.value or Priority.research

    def get_slurm_qos_for_case(self, case_id: str) -> str:
        """Get Quality of service (SLURM QOS) for the case."""
        priority: int = self.get_priority_for_case(case_id=case_id)
        return PRIORITY_TO_SLURM_QOS[priority]

    def get_workflow_manager(self) -> str:
        """Get workflow manager for a given pipeline."""
        return WorkflowManager.Slurm.value

    def get_case_path(self, case_id: str) -> list[Path] | Path:
        """Path to case working directory."""
        raise NotImplementedError

    def get_case_config_path(self, case_id) -> Path:
        """Path to case config file"""
        raise NotImplementedError

    def get_trailblazer_config_path(self, case_id: str) -> Path:
        """Path to Trailblazer job id file"""
        raise NotImplementedError

    def get_sample_name_from_lims_id(self, lims_id: str) -> str:
        """Retrieve sample name provided by customer for specific sample"""
        sample: Sample = self.status_db.get_sample_by_internal_id(internal_id=lims_id)
        return sample.name

    def link_fastq_files(self, case_id: str, dry_run: bool = False) -> None:
        """
        Links fastq files from Housekeeper to case working directory
        """
        raise NotImplementedError

    def get_bundle_deliverables_type(self, case_id: str) -> str | None:
        return None

    @staticmethod
    def get_application_type(sample_obj: Sample) -> str:
        """
        Gets application type for sample. Only application types supported by trailblazer (or other)
        are valid outputs.
        """
        prep_category: str = sample_obj.prep_category
        if prep_category and prep_category.lower() in {
            AnalysisType.TARGETED_GENOME_SEQUENCING,
            AnalysisType.WHOLE_EXOME_SEQUENCING,
            AnalysisType.WHOLE_GENOME_SEQUENCING,
            AnalysisType.WHOLE_TRANSCRIPTOME_SEQUENCING,
        }:
            return prep_category.lower()
        return AnalysisType.OTHER

    def upload_bundle_housekeeper(self, case_id: str, dry_run: bool = False) -> None:
        """Storing bundle data in Housekeeper for CASE_ID"""
        LOG.info(f"Storing bundle data in Housekeeper for {case_id}")
        bundle_data: dict = self.get_hermes_transformed_deliverables(case_id)
        bundle_result: tuple[Bundle, Version] = self.housekeeper_api.add_bundle(
            bundle_data=bundle_data
        )
        if not bundle_result:
            LOG.info("Bundle already added to Housekeeper!")
            raise BundleAlreadyAddedError("Bundle already added to Housekeeper!")
        bundle_object, bundle_version = bundle_result
        if dry_run:
            LOG.info("Dry-run: Housekeeper changes will not be commited")
            LOG.info(
                "The following files would be stored:\n%s",
                "\n".join([f["path"] for f in bundle_data["files"]]),
            )
            return
        self.housekeeper_api.include(bundle_version)
        self.housekeeper_api.add_commit(bundle_object)
        self.housekeeper_api.add_commit(bundle_version)
        LOG.info(
            f"Analysis successfully stored in Housekeeper: {case_id} : {bundle_version.created_at}"
        )

    def upload_bundle_statusdb(self, case_id: str, dry_run: bool = False) -> None:
        """Storing analysis bundle in StatusDB for CASE_ID"""

        LOG.info(f"Storing analysis in StatusDB for {case_id}")
        case_obj: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        analysis_start: dt.datetime = self.get_bundle_created_date(case_id=case_id)
        pipeline_version: str = self.get_pipeline_version(case_id=case_id)
        new_analysis: Case = self.status_db.add_analysis(
            pipeline=self.pipeline,
            version=pipeline_version,
            started_at=analysis_start,
            completed_at=dt.datetime.now(),
            primary=(len(case_obj.analyses) == 0),
        )
        new_analysis.case = case_obj
        if dry_run:
            LOG.info("Dry-run: StatusDB changes will not be commited")
            return
        self.status_db.session.add(new_analysis)
        self.status_db.session.commit()
        LOG.info(f"Analysis successfully stored in StatusDB: {case_id} : {analysis_start}")

    def get_deliverables_file_path(self, case_id: str) -> Path:
        raise NotImplementedError

    def get_analysis_finish_path(self, case_id: str) -> Path:
        raise NotImplementedError

    def add_pending_trailblazer_analysis(self, case_id: str) -> None:
        self.check_analysis_ongoing(case_id=case_id)
        self.trailblazer_api.mark_analyses_deleted(case_id=case_id)
        self.trailblazer_api.add_pending_analysis(
            case_id=case_id,
            email=environ_email(),
            analysis_type=self.get_application_type(
                self.status_db.get_case_by_internal_id(internal_id=case_id).links[0].sample
            ),
            out_dir=self.get_trailblazer_config_path(case_id=case_id).parent.as_posix(),
            config_path=self.get_trailblazer_config_path(case_id=case_id).as_posix(),
            slurm_quality_of_service=self.get_slurm_qos_for_case(case_id=case_id),
            data_analysis=str(self.pipeline),
            ticket=self.status_db.get_latest_ticket_from_case(case_id),
            workflow_manager=self.get_workflow_manager(),
        )

    def get_hermes_transformed_deliverables(self, case_id: str) -> dict:
        return self.hermes_api.create_housekeeper_bundle(
            bundle_name=case_id,
            deliverables=self.get_deliverables_file_path(case_id=case_id),
            pipeline=str(self.pipeline),
            analysis_type=self.get_bundle_deliverables_type(case_id),
            created=self.get_bundle_created_date(case_id),
        ).model_dump()

    def get_bundle_created_date(self, case_id: str) -> dt.datetime:
        return self.get_date_from_file_path(self.get_deliverables_file_path(case_id=case_id))

    def get_pipeline_version(self, case_id: str) -> str:
        """
        Calls the pipeline to get the pipeline version number. If fails, returns a placeholder value instead.
        """
        try:
            self.process.run_command(["--version"])
            return list(self.process.stdout_lines())[0].split()[-1]
        except (Exception, CalledProcessError):
            LOG.warning("Could not retrieve %s workflow version!", self.pipeline)
            return "0.0.0"

    def set_statusdb_action(self, case_id: str, action: str | None, dry_run: bool = False) -> None:
        """
        Set one of the allowed actions on a case in StatusDB.
        """
        if dry_run:
            LOG.info(f"Dry-run: Action {action} would be set for case {case_id}")
            return
        if action in [None, *CASE_ACTIONS]:
            case_obj: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
            case_obj.action = action
            self.status_db.session.commit()
            LOG.info("Action %s set for case %s", action, case_id)
            return
        LOG.warning(
            f"Action '{action}' not permitted by StatusDB and will not be set for case {case_id}"
        )

    def get_analyses_to_clean(self, before: dt.datetime) -> list[Analysis]:
        analyses_to_clean = self.status_db.get_analyses_to_clean(
            pipeline=self.pipeline, before=before
        )
        return analyses_to_clean

    def get_cases_to_analyze(self) -> list[Case]:
        return self.status_db.cases_to_analyze(
            pipeline=self.pipeline, threshold=self.use_read_count_threshold
        )

    def get_cases_to_store(self) -> list[Case]:
        """Return cases where analysis finished successfully,
        and is ready to be stored in Housekeeper."""
        return [
            case
            for case in self.status_db.get_running_cases_in_pipeline(pipeline=self.pipeline)
            if self.trailblazer_api.is_latest_analysis_completed(case_id=case.internal_id)
        ]

    def get_cases_to_qc(self) -> list[Case]:
        """Return cases where analysis finished successfully,
        and is ready for QC metrics checks."""
        return [
            case
            for case in self.status_db.get_running_cases_in_pipeline(pipeline=self.pipeline)
            if self.trailblazer_api.is_latest_analysis_qc(case_id=case.internal_id)
        ]

    def get_sample_fastq_destination_dir(self, case: Case, sample: Sample):
        """Return the path to the FASTQ destination directory."""
        raise NotImplementedError

    def gather_file_metadata_for_sample(self, sample_obj: Sample) -> list[dict]:
        return [
            self.fastq_handler.parse_file_data(file_obj.full_path)
            for file_obj in self.housekeeper_api.files(
                bundle=sample_obj.internal_id, tags=["fastq"]
            )
        ]

    def link_fastq_files_for_sample(
        self, case_obj: Case, sample_obj: Sample, concatenate: bool = False
    ) -> None:
        """
        Link FASTQ files for a sample to working directory.
        If pipeline input requires concatenated fastq, files can also be concatenated
        """
        linked_reads_paths = {1: [], 2: []}
        concatenated_paths = {1: "", 2: ""}
        files: list[dict] = self.gather_file_metadata_for_sample(sample_obj=sample_obj)
        sorted_files = sorted(files, key=lambda k: k["path"])
        fastq_dir = self.get_sample_fastq_destination_dir(case=case_obj, sample=sample_obj)
        fastq_dir.mkdir(parents=True, exist_ok=True)

        for fastq_data in sorted_files:
            fastq_path = Path(fastq_data["path"])
            fastq_name = self.fastq_handler.create_fastq_name(
                lane=fastq_data["lane"],
                flowcell=fastq_data["flowcell"],
                sample=sample_obj.internal_id,
                read=fastq_data["read"],
                undetermined=fastq_data["undetermined"],
                meta=self.get_additional_naming_metadata(sample_obj),
            )
            destination_path: Path = fastq_dir / fastq_name
            linked_reads_paths[fastq_data["read"]].append(destination_path)
            concatenated_paths[
                fastq_data["read"]
            ] = f"{fastq_dir}/{self.fastq_handler.get_concatenated_name(fastq_name)}"

            if not destination_path.exists():
                LOG.info(f"Linking: {fastq_path} -> {destination_path}")
                destination_path.symlink_to(fastq_path)
            else:
                LOG.warning(f"Destination path already exists: {destination_path}")

        if not concatenate:
            return

        LOG.info("Concatenation in progress for sample %s.", sample_obj.internal_id)
        for read, value in linked_reads_paths.items():
            self.fastq_handler.concatenate(linked_reads_paths[read], concatenated_paths[read])
            self.fastq_handler.remove_files(value)

    def get_target_bed_from_lims(self, case_id: str) -> str | None:
        """Get target bed filename from LIMS."""
        case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        sample: Sample = case.links[0].sample
        if sample.from_sample:
            sample: Sample = self.status_db.get_sample_by_internal_id(
                internal_id=sample.from_sample
            )
        target_bed_shortname: str = self.lims_api.capture_kit(lims_id=sample.internal_id)
        if not target_bed_shortname:
            return None
        bed_version: BedVersion | None = self.status_db.get_bed_version_by_short_name(
            bed_version_short_name=target_bed_shortname
        )
        if not bed_version:
            raise CgDataError(f"Bed-version {target_bed_shortname} does not exist")
        return bed_version.filename

    def decompression_running(self, case_id: str) -> None:
        """Check if decompression is running for a case"""
        is_decompression_running: bool = self.prepare_fastq_api.is_spring_decompression_running(
            case_id
        )
        if not is_decompression_running:
            LOG.warning(
                "Decompression can not be started for %s",
                case_id,
            )
            return
        LOG.info(
            "Decompression is running for %s, analysis will be started when decompression is done",
            case_id,
        )
        self.set_statusdb_action(case_id=case_id, action="analyze")

    def decompress_case(self, case_id: str, dry_run: bool) -> None:
        """Decompress all possible fastq files for all samples in a case"""
        LOG.info(
            "The analysis for %s could not start, decompression is needed",
            case_id,
        )
        decompression_possible: bool = (
            self.prepare_fastq_api.can_at_least_one_sample_be_decompressed(case_id)
        )
        if not decompression_possible:
            self.decompression_running(case_id=case_id)
            return
        case_obj: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        link: CaseSample
        any_decompression_started = False
        for link in case_obj.links:
            sample_id: str = link.sample.internal_id
            if dry_run:
                LOG.info(
                    f"This is a dry run, therefore decompression for {sample_id} won't be started"
                )
                continue
            decompression_started: bool = self.prepare_fastq_api.compress_api.decompress_spring(
                sample_id
            )
            if decompression_started:
                any_decompression_started = True

        if not any_decompression_started:
            LOG.warning("Decompression failed to start for %s", case_id)
            return
        if not dry_run:
            self.set_statusdb_action(case_id=case_id, action="analyze")
        LOG.info("Decompression started for %s", case_id)

    def resolve_decompression(self, case_id: str, dry_run: bool) -> None:
        """
        Decompresses a case if needed and adds new fastq files to Housekeeper.
        """
        if self.prepare_fastq_api.is_spring_decompression_needed(case_id):
            self.decompress_case(case_id=case_id, dry_run=dry_run)
            return

        if self.prepare_fastq_api.is_spring_decompression_running(case_id):
            self.set_statusdb_action(case_id=case_id, action=CaseActions.ANALYZE)
            return

        self.prepare_fastq_api.add_decompressed_fastq_files_to_housekeeper(case_id)

    @staticmethod
    def get_date_from_file_path(file_path: Path) -> dt.datetime.date:
        """
        Get date from deliverables path using date created metadata.
        """
        return dt.datetime.fromtimestamp(int(os.path.getctime(file_path)))

    def get_additional_naming_metadata(self, sample_obj: Sample) -> str | None:
        return None

    def get_latest_metadata(self, case_id: str) -> AnalysisModel:
        """Get the latest metadata of a specific case"""

        raise NotImplementedError

    def parse_analysis(
        self, config_raw: dict, qc_metrics_raw: dict, sample_info_raw: dict
    ) -> AnalysisModel:
        """Parses output analysis files"""

        raise NotImplementedError

    def clean_analyses(self, case_id: str) -> None:
        """Add a cleaned at date for all analyses related to a case."""
        analyses: list = self.status_db.get_case_by_internal_id(internal_id=case_id).analyses
        LOG.info(f"Adding a cleaned at date for case {case_id}")
        for analysis_obj in analyses:
            analysis_obj.cleaned_at = analysis_obj.cleaned_at or dt.datetime.now()
            self.status_db.session.commit()

    def clean_run_dir(self, case_id: str, yes: bool, case_path: list[Path] | Path) -> int:
        """Remove workflow run directory."""

        try:
            self.verify_case_path_exists(case_id=case_id)
        except FileNotFoundError:
            self.clean_analyses(case_id)

        if yes or click.confirm(f"Are you sure you want to remove all files in {case_path}?"):
            if case_path.is_symlink():
                LOG.warning(
                    f"Will not automatically delete symlink: {case_path}, delete it manually",
                )
                return EXIT_FAIL

            shutil.rmtree(case_path, ignore_errors=True)
            LOG.info(f"Cleaned {case_path}")
            self.clean_analyses(case_id=case_id)
            return EXIT_SUCCESS

    def _is_flow_cell_check_applicable(self, case_id) -> bool:
        """Returns true if the case is neither down sampled nor external."""
        return not (
            self.status_db.is_case_down_sampled(case_id) or self.status_db.is_case_external(case_id)
        )

    def ensure_flow_cells_on_disk(self, case_id: str) -> None:
        """Check if flow cells are on disk for given case. If not, request flow cells."""
        if not self._is_flow_cell_check_applicable(case_id):
            LOG.info(
                "Flow cell check is not applicable - "
                "the case is either down sampled or external."
            )
            return
        if not self.status_db.are_all_flow_cells_on_disk(case_id=case_id):
            self.status_db.request_flow_cells_for_case(case_id)

    def is_case_ready_for_analysis(self, case_id: str) -> bool:
        if self._is_flow_cell_check_applicable(
            case_id
        ) and not self.status_db.are_all_flow_cells_on_disk(case_id):
            LOG.warning(f"Case {case_id} is not ready - all flow cells not present on disk.")
            return False
        if self.prepare_fastq_api.is_spring_decompression_needed(
            case_id
        ) or self.prepare_fastq_api.is_spring_decompression_running(case_id):
            LOG.warning(f"Case {case_id} is not ready - not all files are decompressed.")
            return False
        return True

    def prepare_fastq_files(self, case_id: str, dry_run: bool) -> None:
        """Retrieves or decompresses fastq files if needed, upon which an AnalysisNotReady error
        is raised."""
        self.ensure_flow_cells_on_disk(case_id)
        self.resolve_decompression(case_id, dry_run=dry_run)
        if not self.is_case_ready_for_analysis(case_id):
            raise AnalysisNotReadyError("FASTQ file are not present for the analysis to start")



import click

from cg.constants.constants import Strandedness

OPTION_FROM_START = click.option(
    "--from-start",
    is_flag=True,
    default=False,
    show_default=True,
    help="Start pipeline from start without resuming execution",
)

OPTION_STRANDEDNESS = click.option(
    "--strandedness",
    type=str,
    default=Strandedness.REVERSE.value,
    show_default=True,
    help="Strandedness: forward, unstranded or reverse (default)",
)

OPTION_REFERENCES = click.option(
    "--genomes_base", type=click.Path(), help="Path to references folder"
)



"""API that handles downsampling of samples."""
import logging
from pathlib import Path

from cg.apps.demultiplex.sample_sheet.validators import validate_sample_id
from cg.exc import DownsampleFailedError
from cg.meta.meta import MetaAPI
from cg.meta.workflow.downsample.downsample import DownsampleWorkflow
from cg.models.cg_config import CGConfig
from cg.models.downsample.downsample_data import DownsampleData
from cg.store.models import Case, CaseSample, Sample
from cg.utils.calculations import multiply_by_million

LOG = logging.getLogger(__name__)


class DownsampleAPI(MetaAPI):
    def __init__(
        self,
        config: CGConfig,
        dry_run: bool = False,
    ):
        """Initialize the API."""
        super().__init__(config)
        self.dry_run: bool = dry_run

    def get_downsample_data(
        self, sample_id: str, number_of_reads: float, case_id: str, case_name: str
    ) -> DownsampleData:
        """Return the DownSampleData.
        Raises:
            DownsampleFailedError
        """
        try:
            return DownsampleData(
                status_db=self.status_db,
                hk_api=self.housekeeper_api,
                sample_id=sample_id,
                number_of_reads=number_of_reads,
                case_id=case_id,
                case_name=case_name,
                out_dir=Path(self.config.downsample_dir),
            )
        except Exception as error:
            raise DownsampleFailedError(repr(error))

    def store_downsampled_sample(self, downsample_data: DownsampleData) -> None:
        """
        Add a downsampled sample entry to StatusDB.
        Raises:
            ValueError
        """
        downsampled_sample: Sample = downsample_data.downsampled_sample
        if self.status_db.sample_with_id_exists(sample_id=downsampled_sample.internal_id):
            raise ValueError(f"Sample {downsampled_sample.internal_id} already exists in StatusDB.")
        LOG.info(
            f"New downsampled sample created: {downsampled_sample.internal_id} from {downsampled_sample.from_sample}"
            f"Application tag set to: {downsampled_sample.application_version.application.tag}"
            f"Customer set to: {downsampled_sample.customer}"
        )
        if not self.dry_run:
            self.status_db.session.add(downsampled_sample)
            self.status_db.session.commit()
            LOG.info(f"Added {downsampled_sample.name} to StatusDB.")

    def store_downsampled_case(self, downsample_data: DownsampleData) -> None:
        """
        Add a down sampled case entry to StatusDB.
        """
        downsampled_case: Case = downsample_data.downsampled_case
        if self.status_db.case_with_name_exists(case_name=downsampled_case.name):
            LOG.info(f"Case with name {downsampled_case.name} already exists in StatusDB.")
            return
        if not self.dry_run:
            self.status_db.session.add(downsampled_case)
            self.status_db.session.commit()
            LOG.info(f"New down sampled case created: {downsampled_case.internal_id}")

    def _link_downsampled_sample_to_case(
        self, downsample_data: DownsampleData, sample: Sample, case: Case
    ) -> None:
        """Create a link between sample and case in statusDB."""
        sample_case_link: CaseSample = self.status_db.relate_sample(
            case=case,
            sample=sample,
            status=downsample_data.get_sample_status(),
        )
        if self.dry_run:
            return
        self.status_db.session.add(sample_case_link)
        self.status_db.session.commit()
        LOG.info(f"Related sample {sample.internal_id} to {case.internal_id}")

    def store_downsampled_sample_case(self, downsample_data: DownsampleData) -> None:
        """
        Add the downsampled sample and case to statusDB and generate the sample case link.
        """
        self.store_downsampled_sample(downsample_data)
        self.store_downsampled_case(downsample_data)
        self._link_downsampled_sample_to_case(
            downsample_data=downsample_data,
            sample=downsample_data.downsampled_sample,
            case=downsample_data.downsampled_case,
        )

    def start_downsample_job(
        self,
        downsample_data: DownsampleData,
    ) -> int:
        """
        Start a down sample job for a sample.
        -Retrieves input directory from the latest version of a sample bundle in Housekeeper
        -Starts a down sample job
        """
        downsample_work_flow = DownsampleWorkflow(
            number_of_reads=multiply_by_million(downsample_data.number_of_reads),
            config=self.config,
            output_fastq_dir=str(downsample_data.fastq_file_output_directory),
            input_fastq_dir=str(downsample_data.fastq_file_input_directory),
            original_sample=downsample_data.original_sample,
            downsampled_sample=downsample_data.downsampled_sample,
            dry_run=self.dry_run,
        )
        return downsample_work_flow.write_and_submit_sbatch_script()

    def downsample_sample(
        self, sample_id: str, case_name: str, case_id: str, number_of_reads: float
    ) -> int | None:
        """Downsample a sample."""
        LOG.info(f"Starting Downsampling for sample {sample_id}.")
        validate_sample_id(sample_id)
        downsample_data: DownsampleData = self.get_downsample_data(
            sample_id=sample_id,
            case_id=case_id,
            number_of_reads=number_of_reads,
            case_name=case_name,
        )
        if self.prepare_fastq_api.is_sample_decompression_needed(
            downsample_data.original_sample.internal_id
        ):
            self.prepare_fastq_api.compress_api.decompress_spring(
                downsample_data.original_sample.internal_id
            )
            return
        LOG.debug("No Decompression needed.")
        self.store_downsampled_sample_case(downsample_data=downsample_data)
        downsample_data.create_down_sampling_working_directory()
        submitted_job: int = self.start_downsample_job(
            downsample_data=downsample_data,
        )
        return submitted_job



from enum import Enum, IntEnum, StrEnum


class Gender(StrEnum):
    FEMALE = "female"
    MALE = "male"
    UNKNOWN = "unknown"
    OTHER = "other"
    MISSING = ""


class PhenotypeStatus(StrEnum):
    UNKNOWN = "unknown"
    UNAFFECTED = "unaffected"
    AFFECTED = "affected"
    MISSING = ""

    def __repr__(self):
        return self.value


class PlinkPhenotypeStatus(IntEnum):
    UNKNOWN = 0
    UNAFFECTED = 1
    AFFECTED = 2


class PlinkGender(str, Enum):
    UNKNOWN = 0
    MALE = 1
    FEMALE = 2

    def __str__(self) -> str:
        return str.__str__(self.value)


class RelationshipStatus(str, Enum):
    HAS_NO_PARENT = 0

    def __str__(self) -> str:
        return str.__str__(self.value)



""" Upload NIPT results via the CLI"""

import logging
import traceback

import click

from cg.constants.nipt import Q30_THRESHOLD
from cg.exc import AnalysisUploadError
from cg.meta.upload.nipt import NiptUploadAPI
from cg.store.models import Analysis

from .ftp import ftp
from .ftp import nipt_upload_case as nipt_upload_ftp_case
from .statina import batch, statina

LOG = logging.getLogger(__name__)


@click.group()
def nipt():
    """Upload NIPT result files"""
    pass


@nipt.command("case")
@click.argument("case_id", required=True)
@click.option("--dry-run", is_flag=True)
@click.option("--force", is_flag=True, help="Force upload of case to databases, despite qc")
@click.pass_context
def nipt_upload_case(context: click.Context, case_id: str | None, dry_run: bool, force: bool):
    """Upload NIPT result files for a case"""

    LOG.info("*** NIPT UPLOAD START ***")

    if context.invoked_subcommand is not None:
        return

    nipt_upload_api: NiptUploadAPI = NiptUploadAPI(context.obj)
    nipt_upload_api.set_dry_run(dry_run=dry_run)
    if force or nipt_upload_api.flowcell_passed_qc_value(
        case_id=case_id, q30_threshold=Q30_THRESHOLD
    ):
        nipt_upload_api.update_analysis_upload_started_date(case_id)
        context.invoke(batch, case_id=case_id, force=force, dry_run=dry_run)
        context.invoke(nipt_upload_ftp_case, case_id=case_id, force=force, dry_run=dry_run)
        nipt_upload_api.update_analysis_uploaded_at_date(case_id)
        LOG.info(f"{case_id}: analysis uploaded!")
    else:
        LOG.error(f"Uploading case failed: {case_id}")
        LOG.error(
            f"Flowcell did not pass one of the following QC parameters:\n"
            f"target_reads={nipt_upload_api.target_reads(case_id=case_id)}, Q30_threshold={Q30_THRESHOLD}"
        )
        raise AnalysisUploadError("Upload failed")


@nipt.command("all")
@click.option("--dry-run", is_flag=True)
@click.pass_context
def nipt_upload_all(context: click.Context, dry_run: bool):
    """Upload NIPT result files for all cases"""

    LOG.info("*** NIPT UPLOAD ALL START ***")

    nipt_upload_api: NiptUploadAPI = NiptUploadAPI(context.obj)
    nipt_upload_api.set_dry_run(dry_run=dry_run)

    all_good = True
    analyses: list[Analysis] = nipt_upload_api.get_all_upload_analyses()
    if not analyses:
        LOG.info("No analyses found to upload")
        return

    for analysis in analyses:
        internal_id = analysis.case.internal_id

        if nipt_upload_api.flowcell_passed_qc_value(
            case_id=internal_id, q30_threshold=Q30_THRESHOLD
        ):
            LOG.info(f"Uploading case: {internal_id}")
            try:
                context.invoke(nipt_upload_case, case_id=internal_id, dry_run=dry_run)
            except AnalysisUploadError:
                LOG.error(traceback.format_exc())
                all_good = False

    if not all_good:
        raise AnalysisUploadError("Some uploads failed")


nipt.add_command(ftp)
nipt.add_command(statina)



"""Code that handles upload CLI commands."""
import logging
import sys
import traceback

import click

from cg.cli.upload.clinical_delivery import auto_fastq, upload_clinical_delivery
from cg.cli.upload.coverage import upload_coverage
from cg.cli.upload.delivery_report import upload_delivery_report_to_scout
from cg.cli.upload.fohm import fohm
from cg.cli.upload.genotype import upload_genotypes
from cg.cli.upload.gens import upload_to_gens
from cg.cli.upload.gisaid import upload_to_gisaid
from cg.cli.upload.mutacc import process_solved, processed_solved
from cg.cli.upload.nipt import nipt
from cg.cli.upload.observations import (
    upload_available_observations_to_loqusdb,
    upload_observations_to_loqusdb,
)
from cg.cli.upload.scout import (
    create_scout_load_config,
    upload_case_to_scout,
    upload_multiqc_to_scout,
    upload_rna_fusion_report_to_scout,
    upload_rna_junctions_to_scout,
    upload_rna_to_scout,
    upload_to_scout,
)
from cg.cli.upload.utils import suggest_cases_to_upload
from cg.cli.upload.validate import validate
from cg.constants import Pipeline
from cg.exc import AnalysisAlreadyUploadedError
from cg.meta.upload.balsamic.balsamic import BalsamicUploadAPI
from cg.meta.upload.mip.mip_dna import MipDNAUploadAPI
from cg.meta.upload.mip.mip_rna import MipRNAUploadAPI
from cg.meta.upload.rnafusion.rnafusion import RnafusionUploadAPI
from cg.meta.upload.upload_api import UploadAPI
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import Case
from cg.utils.click.EnumChoice import EnumChoice

LOG = logging.getLogger(__name__)


@click.group(invoke_without_command=True)
@click.option("-c", "--case", "case_id", help="Upload to all apps")
@click.option(
    "-r",
    "--restart",
    is_flag=True,
    help="Force upload of an analysis that has already been uploaded or marked as started",
)
@click.pass_context
def upload(context: click.Context, case_id: str | None, restart: bool):
    """Upload results from analyses"""

    config_object: CGConfig = context.obj
    upload_api: UploadAPI = MipDNAUploadAPI(config=config_object)  # default upload API

    LOG.info("----------------- UPLOAD -----------------")

    if context.invoked_subcommand is not None:
        context.obj.meta_apis["upload_api"] = upload_api
    elif case_id:  # Provided case ID without a subcommand: upload everything
        try:
            upload_api.analysis_api.status_db.verify_case_exists(case_internal_id=case_id)
            case: Case = upload_api.status_db.get_case_by_internal_id(internal_id=case_id)
            upload_api.verify_analysis_upload(case_obj=case, restart=restart)
        except AnalysisAlreadyUploadedError:
            # Analysis being uploaded or it has been already uploaded
            return

        # Update the upload API based on the data analysis type (MIP-DNA by default)
        # Upload for balsamic, balsamic-umi and balsamic-qc
        if Pipeline.BALSAMIC in case.data_analysis:
            upload_api = BalsamicUploadAPI(config=config_object)
        elif case.data_analysis == Pipeline.RNAFUSION:
            upload_api = RnafusionUploadAPI(config=config_object)
        elif case.data_analysis == Pipeline.MIP_RNA:
            upload_api: UploadAPI = MipRNAUploadAPI(config=config_object)

        context.obj.meta_apis["upload_api"] = upload_api
        upload_api.upload(ctx=context, case=case, restart=restart)
        click.echo(click.style(f"{case_id} analysis has been successfully uploaded", fg="green"))
    else:
        suggest_cases_to_upload(status_db=upload_api.status_db)
        raise click.Abort()


@upload.command("auto")
@click.option("--pipeline", type=EnumChoice(Pipeline), help="Limit to specific pipeline")
@click.pass_context
def upload_all_completed_analyses(context: click.Context, pipeline: Pipeline = None):
    """Upload all completed analyses"""

    LOG.info("----------------- AUTO -----------------")

    status_db: Store = context.obj.status_db

    exit_code = 0
    for analysis_obj in status_db.get_analyses_to_upload(pipeline=pipeline):
        if analysis_obj.case.analyses[0].uploaded_at is not None:
            LOG.warning(
                f"Skipping upload for case {analysis_obj.case.internal_id}. "
                f"It has been already uploaded at {analysis_obj.case.analyses[0].uploaded_at}."
            )
            continue

        case_id = analysis_obj.case.internal_id
        LOG.info(f"Uploading analysis for case: {case_id}")
        try:
            context.invoke(upload, case_id=case_id)
        except Exception:
            LOG.error(f"Case {case_id} upload failed")
            LOG.error(traceback.format_exc())
            exit_code = 1

    sys.exit(exit_code)


upload.add_command(auto_fastq)
upload.add_command(create_scout_load_config)
upload.add_command(fohm)
upload.add_command(nipt)
upload.add_command(process_solved)
upload.add_command(processed_solved)
upload.add_command(upload_available_observations_to_loqusdb)
upload.add_command(upload_case_to_scout)
upload.add_command(upload_clinical_delivery)
upload.add_command(upload_coverage)
upload.add_command(upload_delivery_report_to_scout)
upload.add_command(upload_genotypes)
upload.add_command(upload_multiqc_to_scout)
upload.add_command(upload_observations_to_loqusdb)
upload.add_command(upload_rna_fusion_report_to_scout)
upload.add_command(upload_rna_junctions_to_scout)
upload.add_command(upload_rna_to_scout)
upload.add_command(upload_to_gens)
upload.add_command(upload_to_gisaid)
upload.add_command(upload_to_scout)
upload.add_command(validate)



"""Module for reading and writing comma separated values (CSV) formatted files."""
import csv
import io
from pathlib import Path
from typing import Any

from cg.constants import FileExtensions
from cg.io.validate_path import validate_file_suffix

DELIMITER_TO_SUFFIX = {",": FileExtensions.CSV, "\t": FileExtensions.TSV}


def read_csv(
    file_path: Path, read_to_dict: bool = False, delimiter: str = ","
) -> list[list[str]] | list[dict]:
    """
    Read content in a CSV file to a list of list or list of dict.
    The delimiter parameter can be used to read TSV files.
    """
    validate_file_suffix(path_to_validate=file_path, target_suffix=DELIMITER_TO_SUFFIX[delimiter])
    with open(file_path, "r") as file:
        csv_reader = (
            csv.DictReader(file, delimiter=delimiter)
            if read_to_dict
            else csv.reader(file, delimiter=delimiter)
        )
        return list(csv_reader)


def read_csv_stream(stream: str, delimiter: str = ",") -> list[list[str]]:
    """Read CSV formatted stream."""
    csv_reader = csv.reader(stream.splitlines(), delimiter=delimiter)
    return list(csv_reader)


def write_csv(content: list[list[Any]], file_path: Path, delimiter: str = ",") -> None:
    """Write content to a CSV file."""
    with open(file_path, "w", newline="") as file:
        csv_writer = csv.writer(file, delimiter=delimiter)
        for row in content:
            csv_writer.writerow(row)


def write_csv_stream(content: list[list[Any]], delimiter: str = ",") -> str:
    """Write content to a CSV stream."""
    csv_stream = io.StringIO()
    csv_writer = csv.writer(csv_stream, lineterminator="\n", delimiter=delimiter)
    for row in content:
        csv_writer.writerow(row)
    return csv_stream.getvalue()



""" Test the CLI for run mip-rna """
import logging

from cg.apps.tb import TrailblazerAPI
from cg.cli.workflow.mip_rna.base import run


def test_cg_dry_run(cli_runner, caplog, case_id, email_address, mip_rna_context, mocker):
    """Test print the MIP command to console"""

    caplog.set_level(logging.INFO)
    # GIVEN a cli function
    # WHEN we run a case in dry run mode

    mocker.patch.object(TrailblazerAPI, "is_latest_analysis_ongoing")
    TrailblazerAPI.is_latest_analysis_ongoing.return_value = False

    result = cli_runner.invoke(
        run, ["--dry-run", "--email", email_address, case_id], obj=mip_rna_context
    )

    # THEN the command should be printed
    assert result.exit_code == 0



"""Module for Balsamic PON Analysis API."""

import logging
from pathlib import Path

from cg.constants.constants import Pipeline
from cg.constants.indexes import ListIndexes
from cg.exc import BalsamicStartError
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Case
from cg.utils.utils import build_command_from_dict

LOG = logging.getLogger(__name__)


class BalsamicPonAnalysisAPI(BalsamicAnalysisAPI):
    """Handles communication between Balsamic PON processes and the rest of CG infrastructure."""

    def __init__(
        self,
        config: CGConfig,
        pipeline: Pipeline = Pipeline.BALSAMIC_PON,
    ):
        super().__init__(config=config, pipeline=pipeline)

    def config_case(
        self,
        case_id: str,
        gender: str,
        genome_version: str,
        panel_bed: str,
        pon_cnn: str,
        observations: list[str],
        dry_run: bool = False,
    ) -> None:
        """Creates a config file for BALSAMIC PON analysis."""
        case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        sample_parameters: dict = self.get_sample_params(case_id=case_id, panel_bed=panel_bed)
        if not sample_parameters:
            LOG.error(f"{case_id} has no samples tagged for Balsamic PON analysis")
            raise BalsamicStartError()
        verified_panel_bed: str = self.get_verified_bed(
            panel_bed=panel_bed, sample_data=sample_parameters
        )
        options: list[str] = build_command_from_dict(
            {
                "--case-id": case.internal_id,
                "--analysis-dir": self.root_dir,
                "--fastq-path": self.get_sample_fastq_destination_dir(case),
                "--panel-bed": verified_panel_bed,
                "--genome-version": genome_version,
                "--balsamic-cache": self.balsamic_cache,
                "--version": self.get_next_pon_version(verified_panel_bed),
            }
        )
        parameters: list[str] = ["config", "pon"] + options
        self.process.run_command(parameters=parameters, dry_run=dry_run)

    def get_case_config_path(self, case_id: str) -> Path:
        """Returns the BALSAMIC PON config path."""
        return Path(self.root_dir, case_id, case_id + "_PON.json")

    def get_next_pon_version(self, panel_bed: str) -> str:
        """Returns the next PON version to be generated."""
        latest_pon_file: str = self.get_latest_pon_file(panel_bed=panel_bed)
        next_version = (
            int(Path(latest_pon_file).stem.split("_v")[ListIndexes.LAST.value]) + 1
            if latest_pon_file
            else 1
        )
        return "v" + str(next_version)



"""Functions that deal with modifications of the indexes."""
import logging

from packaging import version
from pydantic import BaseModel

from cg.apps.demultiplex.sample_sheet.models import (
    FlowCellSample,
    FlowCellSampleBcl2Fastq,
    FlowCellSampleBCLConvert,
)
from cg.constants.constants import FileFormat
from cg.constants.sequencing import Sequencers
from cg.io.controller import ReadFile
from cg.models.demultiplex.run_parameters import RunParameters
from cg.resources import VALID_INDEXES_PATH
from cg.utils.utils import get_hamming_distance

LOG = logging.getLogger(__name__)
DNA_COMPLEMENTS: dict[str, str] = {"A": "T", "C": "G", "G": "C", "T": "A"}
INDEX_ONE_PAD_SEQUENCE: str = "AT"
INDEX_TWO_PAD_SEQUENCE: str = "AC"
LONG_INDEX_CYCLE_NR: int = 10
MINIMUM_HAMMING_DISTANCE: int = 3
NEW_CONTROL_SOFTWARE_VERSION: str = "1.7.0"
NEW_REAGENT_KIT_VERSION: str = "1.5"
REAGENT_KIT_PARAMETER_TO_VERSION: dict[str, str] = {"1": "1.0", "3": "1.5"}
SHORT_SAMPLE_INDEX_LENGTH: int = 8


def is_dual_index(index: str) -> bool:
    """Determines if an index in the raw sample sheet is dual index or not."""
    return "-" in index


class Index(BaseModel):
    """Class that represents an index."""

    name: str
    sequence: str


def get_valid_indexes(dual_indexes_only: bool = True) -> list[Index]:
    """Return a list of valid indexes from the valid indexes file."""
    LOG.info(f"Fetch valid indexes from {VALID_INDEXES_PATH}")
    indexes: list[Index] = []
    indexes_csv: list[list[str]] = ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=VALID_INDEXES_PATH
    )
    for row in indexes_csv:
        index_name: str = row[0]
        index_sequence: str = row[1]
        if dual_indexes_only and not is_dual_index(index=index_sequence):
            continue
        indexes.append(Index(name=index_name, sequence=index_sequence))
    return indexes


def get_reagent_kit_version(reagent_kit_version: str) -> str:
    """Derives the reagent kit version from the run parameters."""
    LOG.info(f"Converting reagent kit parameter {reagent_kit_version} to version")
    if reagent_kit_version not in REAGENT_KIT_PARAMETER_TO_VERSION:
        raise SyntaxError(f"Unknown reagent kit version {reagent_kit_version}")

    return REAGENT_KIT_PARAMETER_TO_VERSION[reagent_kit_version]


def get_index_pair(sample: FlowCellSample) -> tuple[str, str]:
    """Returns a sample index separated into index 1 and index 2."""
    if is_dual_index(sample.index):
        index_1, index_2 = sample.index.split("-")
        return index_1.strip(), index_2.strip()
    return sample.index, sample.index2


def is_reverse_complement_needed(run_parameters: RunParameters) -> bool:
    """Return True if the second index requires reverse complement.

    If the run used the new NovaSeq control software version (NEW_CONTROL_SOFTWARE_VERSION)
    and the new reagent kit version (NEW_REAGENT_KIT_VERSION), then it requires reverse complement.
    If the run is NovaSeqX, does not require reverse complement.
    """
    if run_parameters.sequencer == Sequencers.NOVASEQX:
        return False
    control_software_version: str = run_parameters.control_software_version
    reagent_kit_version: str = run_parameters.reagent_kit_version
    LOG.info("Check if run is reverse complement")
    if version.parse(version=control_software_version) < version.parse(
        version=NEW_CONTROL_SOFTWARE_VERSION
    ):
        LOG.warning(
            f"Old software version {control_software_version}, no need for reverse complement"
        )
        return False
    reagent_kit_version: str = get_reagent_kit_version(reagent_kit_version=reagent_kit_version)
    if version.parse(reagent_kit_version) < version.parse(NEW_REAGENT_KIT_VERSION):
        LOG.warning(
            f"Reagent kit version {reagent_kit_version} does not does not need reverse complement"
        )
        return False
    LOG.info("Run is reverse complement")
    return True


def is_padding_needed(index_cycles: int, sample_index_length: int) -> bool:
    """Returns whether a sample needs padding or not given the sample index length.
    A sample needs padding if its adapted index length is shorter than the number of index cycles
    reads stated in the run parameters file of the sequencing. This happens when the sample index
    is 8 nucleotides long and the number of index cycles read is 10 nucleotides.
    """
    return index_cycles == LONG_INDEX_CYCLE_NR and sample_index_length == SHORT_SAMPLE_INDEX_LENGTH


def get_reverse_complement_dna_seq(dna: str) -> str:
    """Generates the reverse complement of a DNA sequence."""
    LOG.debug(f"Reverse complement string {dna}")

    return "".join(DNA_COMPLEMENTS[base] for base in reversed(dna))


def pad_index_one(index_string: str) -> str:
    """Adds bases 'AT' to index one."""
    return index_string + INDEX_ONE_PAD_SEQUENCE


def pad_index_two(index_string: str, reverse_complement: bool) -> str:
    """Adds bases to index two depending on if it should be reverse complement or not."""
    if reverse_complement:
        return INDEX_TWO_PAD_SEQUENCE + index_string
    return index_string + INDEX_TWO_PAD_SEQUENCE


def get_hamming_distance_index_1(sequence_1: str, sequence_2: str) -> int:
    """Get the hamming distance between two index 1 sequences.
    In the case that one sequence is longer than the other, the distance is calculated between
    the shortest sequence and the first segment of equal length of the longest sequence."""
    shortest_index_length: int = min(len(sequence_1), len(sequence_2))
    return get_hamming_distance(
        str_1=sequence_1[:shortest_index_length], str_2=sequence_2[:shortest_index_length]
    )


def get_hamming_distance_index_2(
    sequence_1: str, sequence_2: str, is_reverse_complement: bool
) -> int:
    """Get the hamming distance between two index 2 sequences.
    In the case that one sequence is longer than the other, the distance is calculated between
    the shortest sequence and the last segment of equal length of the longest sequence.
    If the sample requires reverse complement, the calculation is the same as for index 1."""
    shortest_index_length: int = min(len(sequence_1), len(sequence_2))
    return (
        get_hamming_distance(
            str_1=sequence_1[:shortest_index_length], str_2=sequence_2[:shortest_index_length]
        )
        if is_reverse_complement
        else get_hamming_distance(
            str_1=sequence_1[-shortest_index_length:], str_2=sequence_2[-shortest_index_length:]
        )
    )


def update_barcode_mismatch_values_for_sample(
    sample_to_update: FlowCellSampleBCLConvert,
    samples_to_compare_to: list[FlowCellSampleBCLConvert],
    is_reverse_complement: bool,
) -> None:
    """Updates the sample's barcode mismatch values.
    If a sample index has a hamming distance to any other sample lower than the threshold
    (3 nucleotides), the barcode mismatch value for that index is set to zero."""
    index_1_sample_to_update, index_2_sample_to_update = get_index_pair(sample=sample_to_update)
    for sample_to_compare_to in samples_to_compare_to:
        if sample_to_update.sample_id == sample_to_compare_to.sample_id:
            continue
        index_1, index_2 = get_index_pair(sample=sample_to_compare_to)
        if (
            get_hamming_distance_index_1(sequence_1=index_1_sample_to_update, sequence_2=index_1)
            < MINIMUM_HAMMING_DISTANCE
        ):
            LOG.debug(
                f"Turning barcode mismatch for index 1 to 0 for sample {sample_to_update.sample_id}"
            )
            sample_to_update.barcode_mismatches_1 = 0
        if (
            get_hamming_distance_index_2(
                sequence_1=index_2_sample_to_update,
                sequence_2=index_2,
                is_reverse_complement=is_reverse_complement,
            )
            < MINIMUM_HAMMING_DISTANCE
        ):
            LOG.debug(
                f"Turning barcode mismatch for index 2 to 0 for sample {sample_to_update.sample_id}"
            )
            sample_to_update.barcode_mismatches_2 = 0


def pad_and_reverse_complement_sample_indexes(
    sample: FlowCellSample, index_cycles: int, is_reverse_complement: bool
) -> None:
    """Adapts the indexes of sample.
    1. Pad indexes if needed so that all indexes have a length equal to the number of index reads
    2. Takes the reverse complement of index 2 in case of the new NovaSeq software control version
    (1.7) in combination with the new reagent kit (version 1.5).
    3. Assigns the indexes to the sample attributes index and index2."""
    index1, index2 = get_index_pair(sample=sample)
    index_length = len(index1)
    if isinstance(sample, FlowCellSampleBcl2Fastq) and is_padding_needed(
        index_cycles=index_cycles, sample_index_length=index_length
    ):
        LOG.debug("Padding indexes")
        index1 = pad_index_one(index_string=index1)
        index2 = pad_index_two(index_string=index2, reverse_complement=is_reverse_complement)
    LOG.debug(f"Padding not necessary for sample {sample.sample_id}")
    if is_reverse_complement:
        index2 = get_reverse_complement_dna_seq(index2)
    sample.index = index1
    sample.index2 = index2


def update_indexes_for_samples(
    samples: list[FlowCellSampleBCLConvert | FlowCellSampleBcl2Fastq],
    index_cycles: int,
    is_reverse_complement: bool,
) -> None:
    """Updates the values to the fields index1 and index 2 of samples."""
    for sample in samples:
        pad_and_reverse_complement_sample_indexes(
            sample=sample,
            index_cycles=index_cycles,
            is_reverse_complement=is_reverse_complement,
        )



import logging

from cg.store.api.add import AddHandler
from cg.store.api.delete import DeleteDataHandler
from cg.store.api.find_basic_data import FindBasicDataHandler
from cg.store.api.find_business_data import FindBusinessDataHandler
from cg.store.api.status import StatusHandler
from cg.store.api.update import UpdateHandler
from cg.store.database import get_session

LOG = logging.getLogger(__name__)


class CoreHandler(
    AddHandler,
    DeleteDataHandler,
    FindBasicDataHandler,
    FindBusinessDataHandler,
    StatusHandler,
    UpdateHandler,
):
    """Aggregating class for the store api handlers."""

    def __init__(self, session):
        DeleteDataHandler(session)
        FindBasicDataHandler(session)
        FindBusinessDataHandler(session)
        StatusHandler(session)
        UpdateHandler(session)


class Store(CoreHandler):
    def __init__(self):
        self.session = get_session()



""" Module for retrieving flow cells from backup."""
import logging
import subprocess
from pathlib import Path

from housekeeper.store.models import File

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.backup import MAX_PROCESSING_FLOW_CELLS
from cg.constants.constants import FileExtensions, FlowCellStatus
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.constants.indexes import ListIndexes
from cg.constants.symbols import NEW_LINE
from cg.exc import ChecksumFailedError, PdcError, PdcNoFilesMatchingSearchError
from cg.meta.backup.pdc import PdcAPI
from cg.meta.encryption.encryption import EncryptionAPI, SpringEncryptionAPI
from cg.meta.tar.tar import TarAPI
from cg.models import CompressionData
from cg.models.cg_config import PDCArchivingDirectory
from cg.store import Store
from cg.store.models import Flowcell
from cg.utils.time import get_elapsed_time, get_start_time

LOG = logging.getLogger(__name__)


class BackupAPI:
    """Class for retrieving FCs from backup."""

    def __init__(
        self,
        encryption_api: EncryptionAPI,
        pdc_archiving_directory: PDCArchivingDirectory,
        status: Store,
        tar_api: TarAPI,
        pdc_api: PdcAPI,
        flow_cells_dir: str,
        dry_run: bool = False,
    ):
        self.encryption_api = encryption_api
        self.pdc_archiving_directory: PDCArchivingDirectory = pdc_archiving_directory
        self.status: Store = status
        self.tar_api: TarAPI = tar_api
        self.pdc: PdcAPI = pdc_api
        self.flow_cells_dir: str = flow_cells_dir
        self.dry_run: bool = dry_run

    def check_processing(self) -> bool:
        """Check if the processing queue for flow cells is not full."""
        processing_flow_cells_count: int = len(
            self.status.get_flow_cells_by_statuses(flow_cell_statuses=[FlowCellStatus.PROCESSING])
        )
        LOG.debug(f"Processing flow cells: {processing_flow_cells_count}")
        return processing_flow_cells_count < MAX_PROCESSING_FLOW_CELLS

    def get_first_flow_cell(self) -> Flowcell | None:
        """Get the first flow cell from the requested queue."""
        flow_cell: Flowcell | None = self.status.get_flow_cells_by_statuses(
            flow_cell_statuses=[FlowCellStatus.REQUESTED]
        )
        return flow_cell[0] if flow_cell else None

    def fetch_flow_cell(self, flow_cell: Flowcell | None = None) -> float | None:
        """Start fetching a flow cell from backup if possible.

        1. The processing queue is not full.
        2. The requested queue is not emtpy.
        """
        if self.check_processing() is False:
            LOG.info("Processing queue is full")
            return None

        if not flow_cell:
            flow_cell: Flowcell | None = self.get_first_flow_cell()

        if not flow_cell:
            LOG.info("No flow cells requested")
            return None

        flow_cell.status = FlowCellStatus.PROCESSING
        if not self.dry_run:
            self.status.session.commit()
            LOG.info(f"{flow_cell.name}: retrieving from PDC")

        dsmc_output: list[str] = self.query_pdc_for_flow_cell(flow_cell.name)

        archived_key: Path = self.get_archived_encryption_key_path(dsmc_output=dsmc_output)
        archived_flow_cell: Path = self.get_archived_flow_cell_path(dsmc_output=dsmc_output)

        if not self.dry_run:
            return self._process_flow_cell(
                flow_cell=flow_cell,
                archived_key=archived_key,
                archived_flow_cell=archived_flow_cell,
            )

    def _process_flow_cell(
        self, flow_cell: Flowcell, archived_key: Path, archived_flow_cell: Path
    ) -> float:
        """Process a flow cell from backup. Return elapsed time."""
        start_time: float = get_start_time()
        run_dir: Path = Path(self.flow_cells_dir)
        self.retrieve_archived_key(archived_key=archived_key, flow_cell=flow_cell, run_dir=run_dir)
        self.retrieve_archived_flow_cell(
            archived_flow_cell=archived_flow_cell, flow_cell=flow_cell, run_dir=run_dir
        )

        try:
            (
                decrypted_flow_cell,
                encryption_key,
                retrieved_flow_cell,
                retrieved_key,
            ) = self.decrypt_flow_cell(archived_flow_cell, archived_key, run_dir)

            self.extract_flow_cell(decrypted_flow_cell, run_dir)
            self.create_rta_complete(decrypted_flow_cell, run_dir)
            self.unlink_files(
                decrypted_flow_cell, encryption_key, retrieved_flow_cell, retrieved_key
            )
        except subprocess.CalledProcessError as error:
            LOG.error(f"Decryption failed: {error.stderr}")
            if not self.dry_run:
                flow_cell.status: str = FlowCellStatus.REQUESTED
                self.status.session.commit()
            raise error

        return get_elapsed_time(start_time=start_time)

    def unlink_files(
        self,
        decrypted_flow_cell: Path,
        encryption_key: Path,
        retrieved_flow_cell: Path,
        retrieved_key: Path,
    ):
        """Remove files after flow cell has been fetched from PDC."""
        if self.dry_run:
            return
        LOG.debug("Unlink files")
        message = f"{retrieved_flow_cell} not found, skipping removal"
        try:
            retrieved_flow_cell.unlink()
        except FileNotFoundError:
            LOG.info(message)
        try:
            decrypted_flow_cell.unlink()
        except FileNotFoundError:
            LOG.info(message)
        try:
            retrieved_key.unlink()
        except FileNotFoundError:
            LOG.info(message)
        try:
            encryption_key.unlink()
        except FileNotFoundError:
            LOG.info(message)

    @staticmethod
    def create_rta_complete(decrypted_flow_cell: Path, run_dir: Path):
        """Create an RTAComplete.txt file in the flow cell run directory."""
        (
            run_dir / Path(decrypted_flow_cell.stem).stem / DemultiplexingDirsAndFiles.RTACOMPLETE
        ).touch()

    def extract_flow_cell(self, decrypted_flow_cell, run_dir):
        """Extract the flow cell tar archive."""
        extraction_command = self.tar_api.get_extract_file_command(
            input_file=decrypted_flow_cell, output_dir=run_dir
        )
        LOG.debug(f"Extract flow cell command: {extraction_command}")
        self.tar_api.run_tar_command(extraction_command)

    def decrypt_flow_cell(
        self, archived_flow_cell: Path, archived_key: Path, run_dir: Path
    ) -> tuple[Path, Path, Path, Path]:
        """Decrypt the flow cell."""
        retrieved_key: Path = run_dir / archived_key.name
        encryption_key: Path = retrieved_key.with_suffix(FileExtensions.NO_EXTENSION)
        decryption_command: list[str] = self.encryption_api.get_asymmetric_decryption_command(
            input_file=retrieved_key, output_file=encryption_key
        )
        LOG.debug(f"Decrypt key command: {decryption_command}")
        self.encryption_api.run_gpg_command(decryption_command)
        retrieved_flow_cell: Path = run_dir / archived_flow_cell.name
        decrypted_flow_cell: Path = retrieved_flow_cell.with_suffix(FileExtensions.NO_EXTENSION)
        decryption_command: list[str] = self.encryption_api.get_symmetric_decryption_command(
            input_file=retrieved_flow_cell,
            output_file=decrypted_flow_cell,
            encryption_key=encryption_key,
        )
        LOG.debug(f"Decrypt flow cell command: {decryption_command}")
        self.encryption_api.run_gpg_command(decryption_command)
        return decrypted_flow_cell, encryption_key, retrieved_flow_cell, retrieved_key

    def retrieve_archived_key(self, archived_key: Path, flow_cell: Flowcell, run_dir: Path) -> None:
        """Attempt to retrieve an archived key."""
        try:
            self.retrieve_archived_file(
                archived_file=archived_key,
                run_dir=run_dir,
            )
        except PdcError as error:
            LOG.error(f"{flow_cell.name}: key retrieval failed")
            if not self.dry_run:
                flow_cell.status = FlowCellStatus.REQUESTED
                self.status.session.commit()
            raise error

    def retrieve_archived_flow_cell(
        self, archived_flow_cell: Path, flow_cell: Flowcell, run_dir: Path
    ):
        """Attempt to retrieve an archived flow cell."""
        try:
            self.retrieve_archived_file(
                archived_file=archived_flow_cell,
                run_dir=run_dir,
            )
            if not self.dry_run:
                self._set_flow_cell_status_to_retrieved(flow_cell)
        except PdcError as error:
            LOG.error(f"{flow_cell.name}: run directory retrieval failed")
            if not self.dry_run:
                flow_cell.status = FlowCellStatus.REQUESTED
                self.status.session.commit()
            raise error

    def _set_flow_cell_status_to_retrieved(self, flow_cell: Flowcell):
        flow_cell.status = FlowCellStatus.RETRIEVED
        self.status.session.commit()
        LOG.info(f"Status for flow cell {flow_cell.name} set to {flow_cell.status}")

    def query_pdc_for_flow_cell(self, flow_cell_id: str) -> list[str]:
        """Query PDC for a given flow cell id.
        Raise:
            PdcNoFilesMatchingSearchError if no files are found.
        """
        for _, encryption_directory in self.pdc_archiving_directory:
            search_pattern = f"{encryption_directory}*{flow_cell_id}*{FileExtensions.GPG}"
            self.pdc.query_pdc(search_pattern)
            if self.pdc.was_file_found(self.pdc.process.stderr):
                LOG.info(f"Found archived files for PDC query: {search_pattern}")
                return self.pdc.process.stdout.split(NEW_LINE)
            LOG.debug(f"No archived files found for PDC query: {search_pattern}")

        raise PdcNoFilesMatchingSearchError(
            message=f"No flow cell files found at PDC for {flow_cell_id}"
        )

    def retrieve_archived_file(self, archived_file: Path, run_dir: Path) -> None:
        """Retrieve the archived file from PDC to a flow cell runs directory."""
        retrieved_file = Path(run_dir, archived_file.name)
        LOG.debug(f"Retrieving file {archived_file} to {retrieved_file}")
        self.pdc.retrieve_file_from_pdc(
            file_path=str(archived_file), target_path=str(retrieved_file)
        )

    @classmethod
    def get_archived_flow_cell_path(cls, dsmc_output: list[str]) -> Path | None:
        """Get the path of the archived flow cell from a PDC query."""
        flow_cell_line: str = [
            row
            for row in dsmc_output
            if FileExtensions.TAR in row
            and FileExtensions.GZIP in row
            and FileExtensions.GPG in row
        ][ListIndexes.FIRST.value]

        archived_flow_cell = Path(flow_cell_line.split()[4])
        if archived_flow_cell:
            LOG.info(f"Flow cell found: {archived_flow_cell}")
            return archived_flow_cell

    @classmethod
    def get_archived_encryption_key_path(cls, dsmc_output: list[str]) -> Path | None:
        """Get the encryption key for the archived flow cell from a PDC query."""
        encryption_key_line: str = [
            row
            for row in dsmc_output
            if FileExtensions.KEY in row
            and FileExtensions.GPG in row
            and FileExtensions.GZIP not in row
        ][ListIndexes.FIRST.value]

        archived_encryption_key = Path(encryption_key_line.split()[4])
        if archived_encryption_key:
            LOG.info(f"Encryption key found: {archived_encryption_key}")
            return archived_encryption_key


class SpringBackupAPI:
    """Class for handling PDC backup and retrieval of spring compressed fastq files."""

    def __init__(
        self,
        encryption_api: SpringEncryptionAPI,
        pdc_api: PdcAPI,
        hk_api: HousekeeperAPI,
        dry_run: bool = False,
    ):
        self.encryption_api: SpringEncryptionAPI = encryption_api
        self.hk_api: HousekeeperAPI = hk_api
        self.pdc: PdcAPI = pdc_api
        self.dry_run = dry_run

    def encrypt_and_archive_spring_file(self, spring_file_path: Path) -> None:
        """Encrypts and archives a spring file and its decryption key."""
        LOG.debug(f"*** START BACKUP PROCESS OF SPRING FILE {spring_file_path} ***")
        if self.is_compression_ongoing(spring_file_path):
            LOG.info(
                f"Spring (de)compression ongoing, skipping archiving for spring file {spring_file_path}",
            )
            return
        self.encryption_api.cleanup(spring_file_path)
        if self.is_spring_file_archived(spring_file_path):
            LOG.info(
                "Spring file already archived! Spring file will be removed from disk, continuing "
                "to next spring file "
            )
            self.remove_archived_spring_file(spring_file_path)
            return
        try:
            self.encryption_api.spring_symmetric_encryption(spring_file_path)
            self.encryption_api.key_asymmetric_encryption(spring_file_path)
            self.encryption_api.compare_spring_file_checksums(spring_file_path)
            self.pdc.archive_file_to_pdc(
                file_path=str(self.encryption_api.encrypted_spring_file_path(spring_file_path))
            )
            self.pdc.archive_file_to_pdc(
                file_path=str(self.encryption_api.encrypted_key_path(spring_file_path))
            )
            self.mark_file_as_archived(spring_file_path)
            self.encryption_api.cleanup(spring_file_path)
            self.remove_archived_spring_file(spring_file_path)
            LOG.debug("*** ARCHIVING PROCESS COMPLETED SUCCESSFULLY ***")
        except subprocess.CalledProcessError as error:
            LOG.error(f"Encryption failed: {error.stderr}")
            LOG.debug("*** COMMAND PROCESS FAILED! ***")
            self.encryption_api.cleanup(spring_file_path)
        except ChecksumFailedError as error:
            LOG.error(error)
            self.encryption_api.cleanup(spring_file_path)
            LOG.debug("*** CHECKSUM PROCESS FAILED! ***")

    def retrieve_and_decrypt_spring_file(self, spring_file_path: Path) -> None:
        """Retrieves and decrypts a spring file and its decryption key."""
        LOG.info(f"*** START RETRIEVAL PROCESS OF SPRING FILE {spring_file_path} ***")
        try:
            self.pdc.retrieve_file_from_pdc(
                file_path=str(self.encryption_api.encrypted_spring_file_path(spring_file_path)),
            )
            self.pdc.retrieve_file_from_pdc(
                file_path=str(self.encryption_api.encrypted_key_path(spring_file_path)),
            )
            self.encryption_api.key_asymmetric_decryption(spring_file_path)
            self.encryption_api.spring_symmetric_decryption(
                spring_file_path, output_file=spring_file_path
            )
        except subprocess.CalledProcessError as error:
            LOG.error(f"Decryption failed: {error.stderr}")
            LOG.debug("*** RETRIEVAL PROCESS FAILED! ***")
        self.encryption_api.cleanup(spring_file_path)
        LOG.debug("*** RETRIEVAL PROCESS COMPLETED SUCCESSFULLY ***")

    def mark_file_as_archived(self, spring_file_path: Path) -> None:
        """Set the field 'to_archive' of the file in Housekeeper to mark that it has been
        archived to PDC."""
        if self.dry_run:
            LOG.info(f"Dry run, no changes made to {spring_file_path}")
            return
        hk_spring_file: File = self.hk_api.files(path=str(spring_file_path)).first()
        if hk_spring_file:
            LOG.info(f"Setting {spring_file_path} to archived in Housekeeper")
            self.hk_api.set_to_archive(file=hk_spring_file, value=True)
        else:
            LOG.warning(f"Could not find {spring_file_path} on disk")

    def remove_archived_spring_file(self, spring_file_path: Path) -> None:
        """Removes all files related to spring PDC archiving."""
        if not self.dry_run:
            LOG.info(f"Removing spring file {spring_file_path} from disk")
            spring_file_path.unlink()

    def is_to_be_retrieved_and_decrypted(self, spring_file_path: Path) -> bool:
        """Determines if a spring file is archived on PDC and needs to be retrieved and decrypted."""
        spring_file: File = self.hk_api.files(path=str(spring_file_path)).first()
        if spring_file and not spring_file_path.exists():
            return spring_file.to_archive
        return False

    def is_spring_file_archived(self, spring_file_path: Path) -> bool:
        """Checks if a spring file is marked as archived in Housekeeper."""
        spring_file: File = self.hk_api.files(path=str(spring_file_path)).first()
        if spring_file:
            return spring_file.to_archive
        return False

    @staticmethod
    def is_compression_ongoing(spring_file_path: Path) -> bool:
        """Determines if (de)compression of the spring file ongoing."""
        return CompressionData(spring_file_path).pending_exists()



"""Constants related to demultiplexing."""
from enum import Enum, StrEnum
from pathlib import Path

import click

from cg.constants.sequencing import Sequencers


class BclConverter(StrEnum):
    """Define the BCL converter."""

    DRAGEN: str = "dragen"
    BCL2FASTQ: str = "bcl2fastq"
    BCLCONVERT: str = "bcl_convert"


class DemultiplexingDirsAndFiles(StrEnum):
    """Demultiplexing related directories and files."""

    COPY_COMPLETE: str = "CopyComplete.txt"
    DELIVERY: str = "delivery.txt"
    DEMUX_STARTED: str = "demuxstarted.txt"
    DEMUX_COMPLETE: str = "demuxcomplete.txt"
    Hiseq_X_COPY_COMPLETE: str = "copycomplete.txt"
    Hiseq_X_TILE_DIR: str = "l1t11"
    RTACOMPLETE: str = "RTAComplete.txt"
    RUN_PARAMETERS: str = "RunParameters.xml"
    SAMPLE_SHEET_FILE_NAME: str = "SampleSheet.csv"
    UNALIGNED_DIR_NAME: str = "Unaligned"
    BCL2FASTQ_TILE_DIR_PATTERN: str = r"l\dt\d{2}"
    QUEUED_FOR_POST_PROCESSING: str = "post_processing_queued.txt"
    ANALYSIS_COMPLETED: str = "Secondary_Analysis_Complete.txt"
    ANALYSIS: str = "Analysis"
    DATA: str = "Data"
    BCL_CONVERT: str = "BCLConvert"
    FLOW_CELLS_DIRECTORY_NAME: str = "flow_cells"
    DEMULTIPLEXED_RUNS_DIRECTORY_NAME: str = "demultiplexed_runs"
    ILLUMINA_FILE_MANIFEST: str = "Manifest.tsv"
    CG_FILE_MANIFEST: str = "file_manifest.tsv"
    INTER_OP: str = "InterOp"


class RunParametersXMLNodes(StrEnum):
    """Define names of the used XML nodes in run parameters files."""

    # Node names
    APPLICATION: str = ".Application"
    APPLICATION_VERSION: str = ".ApplicationVersion"
    CYCLES: str = "Cycles"
    INDEX_1_NOVASEQ_6000: str = "./IndexRead1NumberOfCycles"
    INDEX_2_NOVASEQ_6000: str = "./IndexRead2NumberOfCycles"
    INDEX_1_NOVASEQ_X: str = "Index1"
    INDEX_2_NOVASEQ_X: str = "Index2"
    INNER_READ: str = ".//Read"
    INSTRUMENT_TYPE: str = ".InstrumentType"
    PLANNED_READS: str = "./PlannedReads"
    READ_1_NOVASEQ_6000: str = "./Read1NumberOfCycles"
    READ_2_NOVASEQ_6000: str = "./Read2NumberOfCycles"
    READ_1_NOVASEQ_X: str = "Read1"
    READ_2_NOVASEQ_X: str = "Read2"
    READ_NAME: str = "ReadName"
    REAGENT_KIT_VERSION: str = "./RfidsInfo/SbsConsumableVersion"

    # Node Values
    NOVASEQ_6000_APPLICATION: str = "NovaSeq Control Software"
    NOVASEQ_X_INSTRUMENT: str = "NovaSeqXPlus"
    UNKNOWN_REAGENT_KIT_VERSION: str = "unknown"


class SampleSheetBcl2FastqSections:
    """Class with all necessary constants for building a NovaSeqX sample sheet."""

    class Settings(Enum):
        HEADER: str = "[Settings]"
        BARCODE_MISMATCH_INDEX1: list[str] = ["BarcodeMismatchesIndex1", "0"]
        BARCODE_MISMATCH_INDEX2: list[str] = ["BarcodeMismatchesIndex2", "0"]

    class Data(Enum):
        HEADER: str = "[Data]"
        FLOW_CELL_ID: str = "FCID"
        LANE: str = "Lane"
        SAMPLE_INTERNAL_ID_BCL2FASTQ: str = "SampleID"
        SAMPLE_REFERENCE: str = "SampleRef"
        INDEX_1: str = "index"
        INDEX_2: str = "index2"
        SAMPLE_NAME: str = "SampleName"
        CONTROL: str = "Control"
        RECIPE: str = "Recipe"
        OPERATOR: str = "Operator"
        SAMPLE_PROJECT_BCL2FASTQ: str = "Project"

        COLUMN_NAMES: list[str] = [
            FLOW_CELL_ID,
            LANE,
            SAMPLE_INTERNAL_ID_BCL2FASTQ,
            SAMPLE_REFERENCE,
            INDEX_1,
            INDEX_2,
            SAMPLE_NAME,
            CONTROL,
            RECIPE,
            OPERATOR,
            SAMPLE_PROJECT_BCL2FASTQ,
        ]


class SampleSheetBCLConvertSections:
    """Class with all necessary constants for building a version 2 sample sheet."""

    class Header(Enum):
        HEADER: str = "[Header]"
        FILE_FORMAT: list[str] = ["FileFormatVersion", "2"]
        RUN_NAME: str = "RunName"
        INSTRUMENT_PLATFORM_TITLE: str = "InstrumentPlatform"
        INSTRUMENT_PLATFORM_VALUE: dict[str, str] = {
            Sequencers.NOVASEQ: "NovaSeq6000",
            Sequencers.NOVASEQX: "NovaSeqXSeries",
        }
        INDEX_ORIENTATION_FORWARD: list[str] = ["IndexOrientation", "Forward"]

    class Reads(StrEnum):
        HEADER: str = "[Reads]"
        READ_CYCLES_1: str = "Read1Cycles"
        READ_CYCLES_2: str = "Read2Cycles"
        INDEX_CYCLES_1: str = "Index1Cycles"
        INDEX_CYCLES_2: str = "Index2Cycles"

    class Settings(Enum):
        HEADER: str = "[BCLConvert_Settings]"
        SOFTWARE_VERSION: list[str] = ["SoftwareVersion", "4.1.7"]
        FASTQ_COMPRESSION_FORMAT: list[str] = ["FastqCompressionFormat", "gzip"]

    class Data(Enum):
        HEADER: str = "[BCLConvert_Data]"
        LANE: str = "Lane"
        SAMPLE_INTERNAL_ID: str = "Sample_ID"
        INDEX_1: str = "Index"
        INDEX_2: str = "Index2"
        OVERRIDE_CYCLES: str = "OverrideCycles"
        ADAPTER_READ_1: str = "AdapterRead1"
        ADAPTER_READ_2: str = "AdapterRead2"
        BARCODE_MISMATCHES_1: str = "BarcodeMismatchesIndex1"
        BARCODE_MISMATCHES_2: str = "BarcodeMismatchesIndex2"

        COLUMN_NAMES: list[str] = [
            LANE,
            SAMPLE_INTERNAL_ID,
            INDEX_1,
            INDEX_2,
            OVERRIDE_CYCLES,
            ADAPTER_READ_1,
            ADAPTER_READ_2,
            BARCODE_MISMATCHES_1,
            BARCODE_MISMATCHES_2,
        ]


OPTION_BCL_CONVERTER = click.option(
    "-b",
    "--bcl-converter",
    type=click.Choice(["bcl2fastq", "dragen"]),
    default=None,
    help="Specify bcl conversion software. Choose between bcl2fastq and dragen. "
    "If not specified, the software will be determined automatically using the sequencer type.",
)


DEMUX_STATS_PATH = {
    "bcl2fastq": {
        "demultiplexing_stats": Path("Stats", "DemultiplexingStats.xml"),
        "conversion_stats": Path("Stats", "ConversionStats.xml"),
        "runinfo": None,
    },
    "dragen": {
        "demultiplexing_stats": Path("Reports", "Demultiplex_Stats.csv"),
        "conversion_stats": Path("Reports", "Demultiplex_Stats.csv"),
        "adapter_metrics_stats": Path("Reports", "Adapter_Metrics.csv"),
        "runinfo": Path("Reports", "RunInfo.xml"),
        "quality_metrics": Path("Reports", "Quality_Metrics.csv"),
    },
}

BCL2FASTQ_METRICS_DIRECTORY_NAME = "Stats"
BCL2FASTQ_METRICS_FILE_NAME = "Stats.json"
DRAGEN_PASSED_FILTER_PCT = 100.00000
FASTQ_FILE_SUFFIXES = [".fastq", ".gz"]
INDEX_CHECK = "indexcheck"
UNDETERMINED = "Undetermined"



from click.testing import CliRunner

from cg.cli.add import add
from cg.constants import EXIT_SUCCESS
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import User


def test_add_user(cli_runner: CliRunner, base_context: CGConfig):
    # GIVEN a database with a customer in it that we can connect the user to
    disk_store: Store = base_context.status_db
    customer_id = "custtest"
    customer = disk_store.add_customer(
        internal_id=customer_id,
        name="Test Customer",
        scout_access=False,
        invoice_address="Street nr, 12345 Uppsala",
        invoice_reference="ABCDEF",
    )
    disk_store.session.add(customer)
    disk_store.session.commit()
    # GIVEN that there is a certain number of users
    user_query = disk_store._get_query(table=User)
    nr_users = user_query.count()

    # WHEN adding a new user
    name, email = "Paul T. Anderson", "paul.anderson@magnolia.com"
    result = cli_runner.invoke(add, ["user", "-c", customer_id, email, name], obj=base_context)

    # THEN exit successfully
    assert result.exit_code == EXIT_SUCCESS
    assert user_query.count() == nr_users + 1



"""Module for deliveries of workflow files"""

import logging
import os
from copy import deepcopy
from pathlib import Path
from typing import Iterable

from housekeeper.store.models import File, Version

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import delivery as constants
from cg.constants.constants import DataDelivery
from cg.exc import MissingFilesError
from cg.store import Store
from cg.store.models import Case, CaseSample, Sample

LOG = logging.getLogger(__name__)


class DeliverAPI:
    """Deliver API for workflows files."""

    def __init__(
        self,
        store: Store,
        hk_api: HousekeeperAPI,
        case_tags: list[set[str]],
        sample_tags: list[set[str]],
        project_base_path: Path,
        delivery_type: str,
        force_all: bool = False,
        ignore_missing_bundles: bool = False,
    ):
        """Initialize a delivery api

        A delivery is made in the context of a ticket id that can be associated to one or many cases.
        Each case can have one or multiple samples linked to them.

        Each delivery is built around case tags and sample tags. All files tagged will the case_tags will be hard linked
        to the inbox of a customer under <ticket>/<case_id>. All files tagged with sample_tags will be linked to
        <ticket>/<case_id>/<sample_id>.
        """
        self.store = store
        self.hk_api = hk_api
        self.project_base_path: Path = project_base_path
        self.case_tags: list[set[str]] = case_tags
        self.all_case_tags: set[str] = {tag for tags in case_tags for tag in tags}
        self.sample_tags: list[set[str]] = sample_tags
        self.customer_id: str = ""
        self.ticket: str = ""
        self.dry_run = False
        self.delivery_type: str = delivery_type
        self.skip_missing_bundle: bool = (
            self.delivery_type in constants.SKIP_MISSING or ignore_missing_bundles
        )
        self.deliver_failed_samples = force_all

    def set_dry_run(self, dry_run: bool) -> None:
        """Update dry run."""
        LOG.info(f"Set dry run to {dry_run}")
        self.dry_run = dry_run

    def deliver_files(self, case_obj: Case):
        """Deliver all files for a case.

        If there are sample tags deliver all files for the samples as well.
        """
        case_id: str = case_obj.internal_id
        case_name: str = case_obj.name
        LOG.debug(
            f"Fetch latest version for case {case_id}",
        )
        last_version: Version = self.hk_api.last_version(bundle=case_id)
        if not last_version:
            if not self.case_tags:
                LOG.info(f"Could not find any version for {case_id}")
            elif not self.skip_missing_bundle:
                raise SyntaxError(f"Could not find any version for {case_id}")
        links: list[CaseSample] = self.store.get_case_samples_by_case_id(case_internal_id=case_id)
        if not links:
            LOG.warning(f"Could not find any samples linked to case {case_id}")
            return
        samples: list[Sample] = [link.sample for link in links]
        self.set_ticket(case_obj.latest_ticket)
        self.set_customer_id(case_obj=case_obj)

        sample_ids: set[str] = {sample.internal_id for sample in samples}

        if self.case_tags:
            self.deliver_case_files(
                case_id=case_id,
                case_name=case_name,
                version=last_version,
                sample_ids=sample_ids,
            )

        if not self.sample_tags:
            return

        link: CaseSample
        for link in links:
            if link.sample.sequencing_qc or self.deliver_failed_samples:
                sample_id: str = link.sample.internal_id
                sample_name: str = link.sample.name
                LOG.debug(f"Fetch last version for sample bundle {sample_id}")
                if self.delivery_type == DataDelivery.FASTQ:
                    last_version: Version = self.hk_api.last_version(bundle=sample_id)
                if not last_version:
                    if self.skip_missing_bundle:
                        LOG.info(f"Could not find any version for {sample_id}")
                        continue
                    raise SyntaxError(f"Could not find any version for {sample_id}")
                self.deliver_sample_files(
                    case_id=case_id,
                    case_name=case_name,
                    sample_id=sample_id,
                    sample_name=sample_name,
                    version_obj=last_version,
                )
                continue
            LOG.warning(
                f"Sample {link.sample.internal_id} did not receive enough reads and will not be delivered"
            )

    def deliver_case_files(
        self, case_id: str, case_name: str, version: Version, sample_ids: set[str]
    ) -> None:
        """Deliver files on case level."""
        LOG.debug(f"Deliver case files for {case_id}")
        # Make sure that the directory exists
        delivery_base: Path = self.create_delivery_dir_path(case_name=case_name)
        LOG.debug(f"Creating project path {delivery_base}")
        if not self.dry_run:
            delivery_base.mkdir(parents=True, exist_ok=True)
        file_path: Path
        number_linked_files: int = 0
        for file_path in self.get_case_files_from_version(version=version, sample_ids=sample_ids):
            # Out path should include customer names
            out_path: Path = delivery_base / file_path.name.replace(case_id, case_name)
            if out_path.exists():
                LOG.warning(f"File {out_path} already exists!")
                continue

            if self.dry_run:
                LOG.info(f"Would hard link file {file_path} to {out_path}")
                number_linked_files += 1
                continue
            LOG.info(f"Hard link file {file_path} to {out_path}")
            try:
                os.link(file_path, out_path)
                number_linked_files += 1
            except FileExistsError:
                LOG.info(f"Path {out_path} exists, skipping")

        LOG.info(f"Linked {number_linked_files} files for case {case_id}")

    def deliver_sample_files(
        self,
        case_id: str,
        case_name: str,
        sample_id: str,
        sample_name: str,
        version_obj: Version,
    ) -> None:
        """Deliver files on sample level."""
        # Make sure that the directory exists
        if self.delivery_type in constants.ONLY_ONE_CASE_PER_TICKET:
            case_name = None
        delivery_base: Path = self.create_delivery_dir_path(
            case_name=case_name, sample_name=sample_name
        )
        LOG.debug(f"Creating project path {delivery_base}")
        if not self.dry_run:
            delivery_base.mkdir(parents=True, exist_ok=True)
        file_path: Path
        number_linked_files_now: int = 0
        number_previously_linked_files: int = 0
        for file_path in self.get_sample_files_from_version(
            version_obj=version_obj, sample_id=sample_id
        ):
            # Out path should include customer names
            file_name: str = file_path.name.replace(sample_id, sample_name)
            if case_name:
                file_name: str = file_name.replace(case_id, case_name)
            out_path: Path = delivery_base / file_name
            if self.dry_run:
                LOG.info(f"Would hard link file {file_path} to {out_path}")
                number_linked_files_now += 1
                continue
            LOG.info(f"Hard link file {file_path} to {out_path}")
            try:
                os.link(file_path, out_path)
                number_linked_files_now += 1
            except FileExistsError:
                LOG.info(
                    f"Warning: Path {out_path} exists, no hard link was made for file {file_name}"
                )
                number_previously_linked_files += 1
        if number_previously_linked_files == 0 and number_linked_files_now == 0:
            raise MissingFilesError(f"No files were linked for sample {sample_name}")

        LOG.info(
            f"There were {number_previously_linked_files} previously linked files and {number_linked_files_now} were linked for sample {sample_id}, case {case_id}"
        )

    def get_case_files_from_version(self, version: Version, sample_ids: set[str]) -> Iterable[Path]:
        """Fetch all case files from a version that are tagged with any of the case tags."""

        if not version:
            LOG.warning("Version is None, cannot get files")
            return []

        if not version.files:
            LOG.warning(f"No files associated with Housekeeper version {version.id}")
            return []

        version_file: File
        for version_file in version.files:
            if not self.include_file_case(file=version_file, sample_ids=sample_ids):
                LOG.debug(f"Skipping file {version_file.path}")
                continue
            yield Path(version_file.full_path)

    def get_sample_files_from_version(self, version_obj: Version, sample_id: str) -> Iterable[Path]:
        """Fetch all files for a sample from a version that are tagged with any of the sample
        tags."""
        file_obj: File
        for file_obj in version_obj.files:
            if not self.include_file_sample(file_obj, sample_id=sample_id):
                continue
            yield Path(file_obj.full_path)

    def include_file_case(self, file: File, sample_ids: set[str]) -> bool:
        """Check if file should be included in case bundle.

        At least one tag should match between file and tags.
        Do not include files with sample tags.
        """
        file_tags = {tag.name for tag in file.tags}
        if self.all_case_tags.isdisjoint(file_tags):
            LOG.debug("No tags are matching")
            return False

        LOG.debug(f"Found file tags {', '.join(file_tags)}")

        # Check if any of the sample tags exist
        if sample_ids.intersection(file_tags):
            LOG.debug(f"Found sample tag, skipping {file.path}")
            return False

        # Check if any of the file tags matches the case tags
        tags: set[str]
        for tags in self.case_tags:
            LOG.debug(f"check if {tags} is a subset of {file_tags}")
            if tags.issubset(file_tags):
                return True
        LOG.debug(f"Could not find any tags matching file {file.path} with tags {file_tags}")

        return False

    def include_file_sample(self, file_obj: File, sample_id: str) -> bool:
        """Check if file should be included in sample bundle.

        At least one tag should match between file and tags.
        Only include files with sample tag.

        For fastq delivery we know that we want to deliver all files of bundle.
        """
        file_tags = {tag.name for tag in file_obj.tags}
        tags: set[str]
        # Check if any of the file tags matches the sample tags
        for tags in self.sample_tags:
            working_copy = deepcopy(tags)
            if self.delivery_type != "fastq":
                working_copy.add(sample_id)
            if working_copy.issubset(file_tags):
                return True

        return False

    def _set_customer_id(self, customer_id: str) -> None:
        LOG.info(f"Setting customer_id to {customer_id}")
        self.customer_id = customer_id

    def set_customer_id(self, case_obj: Case) -> None:
        """Set the customer_id for this upload"""
        self._set_customer_id(case_obj.customer.internal_id)

    def set_ticket(self, ticket: str) -> None:
        """Set the ticket for this upload"""
        LOG.info(f"Setting ticket to {ticket}")
        self.ticket = ticket

    def create_delivery_dir_path(self, case_name: str = None, sample_name: str = None) -> Path:
        """Create a path for delivering files.

        Note that case name and sample name needs to be the identifiers sent from customer.
        """
        delivery_path: Path = Path(
            self.project_base_path, self.customer_id, constants.INBOX_NAME, self.ticket
        )
        if case_name:
            delivery_path = delivery_path / case_name
        if sample_name:
            delivery_path = delivery_path / sample_name

        return delivery_path

    @staticmethod
    def get_delivery_scope(delivery_arguments: set[str]) -> tuple[bool, bool]:
        """Returns the scope of the delivery, ie whether sample and/or case files were delivered."""
        case_delivery: bool = False
        sample_delivery: bool = False
        for delivery in delivery_arguments:
            if (
                constants.PIPELINE_ANALYSIS_TAG_MAP[delivery]["sample_tags"]
                and delivery in constants.ONLY_ONE_CASE_PER_TICKET
            ):
                sample_delivery = True
            if constants.PIPELINE_ANALYSIS_TAG_MAP[delivery]["case_tags"]:
                case_delivery = True
        return sample_delivery, case_delivery



from pathlib import Path

import pytest

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.housekeeper.models import InputBundle
from cg.apps.tb import TrailblazerAPI
from cg.constants import Pipeline
from cg.meta.compress import CompressAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.meta.workflow.mip_rna import MipRNAAnalysisAPI
from cg.meta.workflow.prepare_fastq import PrepareFastqAPI
from cg.models.cg_config import CGConfig
from cg.store.api.find_business_data import FindBusinessDataHandler
from cg.store.api.status import StatusHandler
from cg.store.models import Case
from tests.store_helpers import StoreHelpers


@pytest.fixture
def mip_dna_config_path() -> str:
    return "tests/fixtures/apps/mip/dna/case_config.yaml"


@pytest.fixture
def mip_case_id() -> str:
    return "yellowhog"


@pytest.fixture
def mip_case_id_non_existing() -> str:
    return "this_case_id_does_not_exist"


@pytest.fixture
def mip_case_ids(mip_case_id: str) -> dict:
    """Dictionary of case ids, connected samples, their name and if they should fail (textbook or not)"""

    return {
        mip_case_id: {
            "textbook": True,
            "name": "F0000001",
            "internal_id": "ACC00000",
        },
        "purplesnail": {
            "textbook": False,
            "name": "F0000003",
            "internal_id": "ACC00000",
        },
        "bluezebra": {
            "textbook": True,
            "name": "F0000002",
            "internal_id": "ACC00000",
        },
    }


@pytest.fixture
def mip_case_dirs(mip_case_ids: dict, project_dir: Path) -> dict:
    """Create case directories and return a dictionary of the tmpdir paths"""

    mip_case_dirs = {}

    for case in mip_case_ids:
        case_dir = Path(project_dir / "cases" / case)
        case_dir.mkdir(exist_ok=True, parents=True)
        mip_case_dirs[case] = case_dir

    return mip_case_dirs


@pytest.fixture
def mip_hermes_dna_deliverables_response_data(
    create_multiqc_html_file,
    create_multiqc_json_file,
    mip_case_id,
    timestamp_yesterday,
) -> InputBundle:
    return InputBundle(
        **{
            "files": [
                {
                    "path": create_multiqc_json_file.as_posix(),
                    "tags": ["multiqc-json", mip_case_id, "mip-dna"],
                },
                {
                    "path": create_multiqc_html_file.as_posix(),
                    "tags": ["multiqc-html", mip_case_id, "mip-dna"],
                },
            ],
            "created": timestamp_yesterday,
            "name": mip_case_id,
        }
    )


@pytest.fixture
def mip_rna_context(
    cg_context: CGConfig,
    analysis_family_single_case: dict,
    helpers: StoreHelpers,
    apptag_rna: str,
    case_id: str,
    housekeeper_api: HousekeeperAPI,
    tb_api,
) -> CGConfig:
    cg_context.housekeeper_api_ = housekeeper_api
    cg_context.trailblazer_api_ = tb_api
    analysis_family_single_case["data_analysis"] = str(Pipeline.MIP_RNA)
    if not cg_context.status_db.get_case_by_internal_id(internal_id=case_id):
        helpers.ensure_case_from_dict(
            cg_context.status_db, case_info=analysis_family_single_case, app_tag=apptag_rna
        )
    cg_context.meta_apis["analysis_api"] = MipRNAAnalysisAPI(cg_context)
    return cg_context


@pytest.fixture
def mip_dna_context(
    cg_context: CGConfig,
    helpers: StoreHelpers,
    mip_case_ids: dict,
    real_housekeeper_api: HousekeeperAPI,
    tb_api,
) -> CGConfig:
    _store = cg_context.status_db
    cg_context.housekeeper_api_ = real_housekeeper_api
    cg_context.trailblazer_api_ = tb_api
    mip_analysis_api = MipDNAAnalysisAPI(config=cg_context)

    # Add apptag to db
    helpers.ensure_application_version(store=_store, application_tag="WGSA", prep_category="wgs")

    # Add sample, cases and relationships to db

    for case_id in mip_case_ids:
        if not _store.get_case_by_internal_id(internal_id=case_id):
            case_obj = helpers.add_case(
                store=_store,
                data_analysis=Pipeline.MIP_DNA,
                internal_id=case_id,
                name=mip_case_ids[case_id]["name"],
            )
            sample = helpers.add_sample(
                store=_store,
                customer_id="cust000",
                application_tag="WGSA",
                application_type="wgs",
                gender="unknown",
            )
            helpers.add_relationship(store=_store, sample=sample, case=case_obj, status="affected")
    cg_context.meta_apis["analysis_api"] = mip_analysis_api
    return cg_context


def setup_mocks(
    mocker,
    can_at_least_one_sample_be_decompressed: bool = False,
    case_to_analyze: Case = None,
    decompress_spring: bool = False,
    has_latest_analysis_started: bool = False,
    is_spring_decompression_needed: bool = False,
    is_spring_decompression_running: bool = False,
) -> None:
    """Helper function to setup the necessary mocks for the decompression logics."""
    mocker.patch.object(StatusHandler, "cases_to_analyze")
    StatusHandler.cases_to_analyze.return_value = [case_to_analyze]

    mocker.patch.object(PrepareFastqAPI, "is_spring_decompression_needed")
    PrepareFastqAPI.is_spring_decompression_needed.return_value = is_spring_decompression_needed

    mocker.patch.object(TrailblazerAPI, "has_latest_analysis_started")
    TrailblazerAPI.has_latest_analysis_started.return_value = has_latest_analysis_started

    mocker.patch.object(PrepareFastqAPI, "can_at_least_one_sample_be_decompressed")
    PrepareFastqAPI.can_at_least_one_sample_be_decompressed.return_value = (
        can_at_least_one_sample_be_decompressed
    )

    mocker.patch.object(CompressAPI, "decompress_spring")
    CompressAPI.decompress_spring.return_value = decompress_spring

    mocker.patch.object(PrepareFastqAPI, "is_spring_decompression_running")
    PrepareFastqAPI.is_spring_decompression_running.return_value = is_spring_decompression_running

    mocker.patch.object(PrepareFastqAPI, "add_decompressed_fastq_files_to_housekeeper")
    PrepareFastqAPI.add_decompressed_fastq_files_to_housekeeper.return_value = None

    mocker.patch.object(MipDNAAnalysisAPI, "get_panel_bed")
    MipDNAAnalysisAPI.get_panel_bed.return_value = "a_string"

    mocker.patch.object(FindBusinessDataHandler, "are_all_flow_cells_on_disk")
    FindBusinessDataHandler.are_all_flow_cells_on_disk.return_value = True


@pytest.fixture
def scout_panel_output() -> str:
    return """##genome_build=37
##gene_panel=OMIM-AUTO,version=22.0,updated_at=2023-10-17,display_name=OMIM-AUTO
##gene_panel=PANELAPP-GREEN,version=6.0,updated_at=2023-10-19,display_name=PanelApp Green Genes
##contig=1
##contig=2
##contig=3
##contig=4
##contig=5
##contig=6
##contig=7
##contig=8
##contig=9
##contig=10
##contig=X
##contig=Y
##contig=MT
#chromosome	gene_start	gene_stop	hgnc_id	hgnc_symbol
1	568915	569121	44571	MTATP8P1
1	860260	879955	28706	SAMD11
2       162164549       162268228       16889   PSMD14
2       162272605       162282381       11590   TBR1
3       120315156       120321347       7699    NDUFB4
3       120347020       120401418       4892    HGD
4       185570767       185616117       26575   PRIMPOL
4       185676749       185747972       3569    ACSL1
5       473425  524447  11073   SLC9A3
5       892758  919472  12307   TRIP13
6       18224099        18265054        2768    DEK
6       19837617        19840915        5363    ID4
7       44256749        44374176        1461    CAMK2B
7       44646171        44748665        8124    OGDH
8       22993101        23021543        11907   TNFRSF10D
8       23047965        23082639        11904   TNFRSF10A
9       35099888        35103154        14559   STOML2
9       35161999        35405335        12566   UNC13B
10      23481256        23483181        23734   PTF1A
10      25305587        25315593        26160   THNSL1
X       154719776       154899605       18308   TMLHE
X       155110956       155173433       11486   VAMP7
Y       6778727 6959724 18502   TBL1Y
Y       14813160        14972764        12633   USP9Y
MT      577     647     7481    MT-TF
MT      648     1601    7470    MT-RNR1"""



from sqlalchemy import create_engine, inspect
from sqlalchemy.engine.base import Engine
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.orm import Session, scoped_session, sessionmaker

from cg.exc import CgError
from cg.store.models import Model

SESSION: scoped_session | None = None
ENGINE: Engine | None = None


def initialize_database(db_uri: str) -> None:
    """Initialize the SQLAlchemy engine and session for status db."""
    global SESSION, ENGINE
    ENGINE = create_engine(db_uri, pool_pre_ping=True, future=True)
    session_factory = sessionmaker(ENGINE)
    SESSION = scoped_session(session_factory)


def get_session() -> Session:
    """Get a SQLAlchemy session with a connection to status db."""
    if not SESSION:
        raise CgError("Database not initialised")
    return SESSION


def get_scoped_session_registry() -> scoped_session | None:
    """Get the scoped session registry for status db."""
    return SESSION


def get_engine() -> Engine:
    """Get the SQLAlchemy engine with a connection to status db."""
    if not ENGINE:
        raise CgError("Database not initialised")
    return ENGINE


def create_all_tables() -> None:
    """Create all tables in status db."""
    session: Session = get_session()
    Model.metadata.create_all(bind=session.get_bind())


def drop_all_tables() -> None:
    """Drop all tables in status db."""
    session: Session = get_session()
    Model.metadata.drop_all(bind=session.get_bind())


def get_tables() -> list[str]:
    """Get a list of all tables in status db."""
    engine: Engine = get_engine()
    inspector: Inspector = inspect(engine)
    return inspector.get_table_names()



""" Module to group PDC related commands """

import logging
from pathlib import Path
from subprocess import CalledProcessError

import psutil

from cg.constants.pdc import DSMCParameters
from cg.constants.process import EXIT_WARNING
from cg.exc import (
    DsmcAlreadyRunningError,
    FlowCellAlreadyBackedUpError,
    FlowCellEncryptionError,
    PdcError,
)
from cg.meta.encryption.encryption import FlowCellEncryptionAPI
from cg.store import Store
from cg.store.models import Flowcell
from cg.utils import Process

LOG = logging.getLogger(__name__)

SERVER = "hasta"
NO_FILE_FOUND_ANSWER = "ANS1092W"
MAX_NR_OF_DSMC_PROCESSES: int = 3


class PdcAPI:
    """Group PDC related commands"""

    def __init__(self, binary_path: str, dry_run: bool = False):
        self.process: Process = Process(binary=binary_path)
        self.dry_run: bool = dry_run

    @classmethod
    def validate_is_dsmc_running(cls) -> bool:
        """Check if a Dsmc process is already running on the system.
        Raises:
            Exception: for all non-exit exceptions.
        """
        is_dsmc_running: bool = False
        dsmc_process_count: int = 0
        try:
            for process in psutil.process_iter():
                if "dsmc" == process.name():
                    dsmc_process_count += 1
        except Exception as error:
            LOG.debug(f"{error}")
        if dsmc_process_count >= MAX_NR_OF_DSMC_PROCESSES:
            is_dsmc_running = True
            LOG.debug("Too many Dsmc processes are already running")
        return is_dsmc_running

    def archive_file_to_pdc(self, file_path: str) -> None:
        """Archive a file by storing it on PDC."""
        if not self.dry_run:
            self.run_dsmc_command(command=DSMCParameters.ARCHIVE_COMMAND + [file_path])

    def query_pdc(self, search_pattern: str) -> None:
        """Query PDC based on a given search pattern."""
        command: list = DSMCParameters.QUERY_COMMAND.copy()
        command.append(search_pattern)
        self.run_dsmc_command(command=command)

    def retrieve_file_from_pdc(self, file_path: str, target_path: str = None) -> None:
        """Retrieve a file from PDC"""
        command: list = DSMCParameters.RETRIEVE_COMMAND.copy()
        command.append(file_path)
        if target_path:
            command.append(target_path)
        self.run_dsmc_command(command=command)

    def run_dsmc_command(self, command: list) -> None:
        """Runs a DSMC command.
        Raises:
            PdcError when unable to process command.
        """
        LOG.debug("Starting DSMC command:")
        LOG.debug(f"{self.process.binary} {' '.join(command)}")
        try:
            self.process.run_command(parameters=command, dry_run=self.dry_run)
        except CalledProcessError as error:
            if error.returncode == EXIT_WARNING:
                LOG.warning(f"{error}")
                return
            raise PdcError(message=f"{error}") from error

    def validate_is_flow_cell_backup_possible(
        self, db_flow_cell: Flowcell, flow_cell_encryption_api: FlowCellEncryptionAPI
    ) -> bool:
        """Check if back-up of flow cell is possible.
        Raises:
            DsmcAlreadyRunningError if there is already a Dsmc process ongoing.
            FlowCellAlreadyBackupError if flow cell is already backed up.
            FlowCellEncryptionError if encryption is not complete.
        """
        if self.validate_is_dsmc_running():
            raise DsmcAlreadyRunningError("Too many Dsmc processes are already running")
        if db_flow_cell and db_flow_cell.has_backup:
            raise FlowCellAlreadyBackedUpError(
                f"Flow cell: {db_flow_cell.name} is already backed-up"
            )
        if not flow_cell_encryption_api.complete_file_path.exists():
            raise FlowCellEncryptionError(
                f"Flow cell: {flow_cell_encryption_api.flow_cell.id} encryption process is not complete"
            )
        LOG.debug("Flow cell can be backed up")

    def backup_flow_cell(
        self, files_to_archive: list[Path], store: Store, db_flow_cell: Flowcell
    ) -> None:
        """Back-up flow cell files."""
        for encrypted_file in files_to_archive:
            if not self.dry_run:
                self.archive_file_to_pdc(file_path=encrypted_file.as_posix())
        if not self.dry_run:
            store.update_flow_cell_has_backup(flow_cell=db_flow_cell, has_backup=True)
            LOG.info(f"Flow cell: {db_flow_cell.name} has been backed up")

    def start_flow_cell_backup(
        self,
        db_flow_cell: Flowcell,
        flow_cell_encryption_api: FlowCellEncryptionAPI,
        status_db: Store,
    ) -> None:
        """Check if back-up of flow cell is possible and if so starts it."""
        self.validate_is_flow_cell_backup_possible(
            db_flow_cell=db_flow_cell, flow_cell_encryption_api=flow_cell_encryption_api
        )
        self.backup_flow_cell(
            files_to_archive=[
                flow_cell_encryption_api.final_passphrase_file_path,
                flow_cell_encryption_api.encrypted_gpg_file_path,
            ],
            store=status_db,
            db_flow_cell=db_flow_cell,
        )

    @staticmethod
    def was_file_found(dsmc_output: str) -> bool:
        """Check if file was found in PDC."""
        return NO_FILE_FOUND_ANSWER not in dsmc_output



"""ObservationsInputFiles test fixtures."""
from pathlib import Path

import pytest

from cg.models.observations.input_files import (
    BalsamicObservationsInputFiles,
    MipDNAObservationsInputFiles,
)


@pytest.fixture(name="observations_input_files_raw")
def observations_input_files_raw(case_id: str, filled_file: Path) -> dict:
    """Return raw observations input files for rare diseases."""
    return {
        "family_ped_path": filled_file,
        "profile_vcf_path": filled_file,
        "snv_vcf_path": filled_file,
        "sv_vcf_path": None,
    }


@pytest.fixture(name="observations_input_files")
def observations_input_files(
    observations_input_files_raw: dict,
) -> MipDNAObservationsInputFiles:
    """Return raw observations input files for rare diseases WES analysis."""
    return MipDNAObservationsInputFiles(**observations_input_files_raw)


@pytest.fixture(name="balsamic_observations_input_files_raw")
def balsamic_observations_input_files_raw(case_id: str, filled_file: Path) -> dict:
    """Return raw observations input files for cancer."""
    return {
        "snv_germline_vcf_path": filled_file,
        "snv_vcf_path": filled_file,
        "sv_germline_vcf_path": filled_file,
        "sv_vcf_path": filled_file,
    }


@pytest.fixture(name="balsamic_observations_input_files")
def balsamic_observations_input_files(
    balsamic_observations_input_files_raw: dict,
) -> BalsamicObservationsInputFiles:
    """Return raw observations input files for cancer WGS analysis."""
    return BalsamicObservationsInputFiles(**balsamic_observations_input_files_raw)



import pytest
from sqlalchemy import update

from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import Application
from tests.store.api.conftest import (
    re_sequenced_sample_store,
    store_failing_sequencing_qc,
)


@pytest.fixture
def nipt_upload_api_context(cg_context: CGConfig, re_sequenced_sample_store: Store) -> CGConfig:
    cg_context.status_db_ = re_sequenced_sample_store

    return cg_context


@pytest.fixture
def nipt_upload_api_failed_fc_context(
    nipt_upload_api_context: CGConfig, sample_id: str, store_failing_sequencing_qc: Store
) -> CGConfig:
    nipt_upload_api_failed_fc_context: CGConfig = nipt_upload_api_context
    nipt_upload_api_failed_fc_context.status_db_ = store_failing_sequencing_qc
    status_db = nipt_upload_api_context.status_db

    application = status_db.get_sample_by_internal_id(
        internal_id=sample_id
    ).application_version.application
    status_db.session.execute(
        update(Application).where(Application.id == application.id).values({"target_reads": 20})
    )

    return nipt_upload_api_failed_fc_context
